"""Sets up logging, based on dantro's logging features"""

import os

import coloredlogs as _coloredlogs
from dantro.logging import REMARK as _DEFAULT_LOG_LEVEL
from dantro.logging import getLogger as _getLogger

_log = _getLogger("utopya")
"""The utopya root logger"""

_DEFAULT_LOG_FORMAT = "%(levelname)-8s %(module)-16s  %(message)s"
"""The default logging format"""

# Add colour logging to the root logger
# See API reference:  https://coloredlogs.readthedocs.io/en/latest/api.html
_coloredlogs.install(
    logger=_log,
    level=_DEFAULT_LOG_LEVEL,
    fmt=_DEFAULT_LOG_FORMAT,
    level_styles=dict(
        trace=dict(faint=True),
        debug=dict(faint=True),
        remark=dict(color=246),  # grey
        note=dict(color="cyan"),
        info=dict(bright=True),
        progress=dict(color="green"),
        caution=dict(color=202),  # orange
        hilight=dict(color="yellow", bold=True),
        success=dict(color="green", bold=True),
        warning=dict(color=202, bold=True),  # orange
        error=dict(color="red"),
        critical=dict(color="red", bold=True),
    ),
    field_styles=dict(
        levelname=dict(bold=True, faint=True), module=dict(faint=True)
    ),
)
_log.debug("Logging configured.")



"""The :py:mod:`utopya` package provides a simulation management and
evaluation framework with the following components:

.. default-domain:: utopya

- A registry framework for models (:py:mod:`.model_registry`)
  and projects (:py:class:`~.project_registry.ProjectRegistry`)
- A configuration manager and simulation runner,
  the :py:class:`~.multiverse.Multiverse`:

  - Contains a multi-level configuration interface
  - Parallel simulation execution via
    :py:class:`~.workermanager.WorkerManager`.
  - A :py:mod:`parameter validation framework <.parameter>`

- Coupling to the :py:mod:`dantro` data evaluation pipeline, integrated via
  :py:mod:`utopya.eval`:

  - Custom data :py:mod:`groups <.eval.groups>`
    and :py:mod:`containers <.eval.containers>`
  - A :py:class:`.eval.plotmanager.PlotManager` that takes into account
    project- or model-specific plot function definitions.

- The :py:class:`.model.Model` abstraction which allows convenient interactive
  work with utopya and registered models.

  - The :py:class:`.testtools.ModelTest` class, containing specializations
    that make it more convenient to implement model tests using utopya.

- Batch simulation running and evaluation via :py:mod:`.batch`

For a real-world example of how utopya can be integrated, have a look at the
`Utopia modelling framework <https://utopia-project.org>`_ which uses
utopya as its frontend.
For model implementations, the :py:mod:`utopya_backend` package can assist in
building Python-based models that use :py:mod:`utopya` as a frontend.

Also visit :ref:`the user manual front page <welcome>` for more information.
"""

# Set up logging (needs to happen first), then make the most important parts of
# the utopya interface available

from ._logging import _DEFAULT_LOG_LEVEL, _getLogger
from .eval import DataManager, MultiverseGroup, PlotManager, UniverseGroup
from .model import Model
from .model_registry import MODELS
from .multiverse import DistributedMultiverse, FrozenMultiverse, Multiverse
from .project_registry import PROJECTS
from .testtools import ModelTest

__version__ = "1.3.0b3"
"""The :py:mod:`utopya` package version"""



"""Test the data container"""

import numpy as np
import pytest
from pkg_resources import resource_filename

from utopya.eval.containers import GridDC
from utopya.testtools import ModelTest

from .. import ADVANCED_MODEL
from .._fixtures import *

# -----------------------------------------------------------------------------


def test_griddc():
    """Test the functionality of the GridDC."""
    ### 1d data . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
    # Create some one dimensional test data of the form
    # [ 0  1  2  3  4  5 ]
    # Data in columns represents the time and data in rows the grid data
    data_1d = np.arange(6)

    # Create a GridDC and assert that the shape is correct
    attrs_1d = dict(content="grid", grid_shape=(2, 3), index_order="F")
    extra_attrs = dict(foo="bar")
    gdc_1d = GridDC(
        name="data_1d", data=data_1d, attrs=dict(**attrs_1d, **extra_attrs)
    )

    assert gdc_1d.shape == (2, 3)
    assert "x" in gdc_1d.dims
    assert "y" in gdc_1d.dims
    assert gdc_1d.attrs == dict(**attrs_1d, **extra_attrs)  # carried through

    for i in range(attrs_1d["grid_shape"][0]):
        assert i == gdc_1d.data.coords["x"][i]

    for i in range(attrs_1d["grid_shape"][1]):
        assert i == gdc_1d.data.coords["y"][i]

    # Assert that the data is correct and have the form:
    assert (gdc_1d == np.array([[0, 2, 4], [1, 3, 5]])).all()

    # The format string should contains dimension information
    assert "x: 2" in gdc_1d._format_info()
    assert "y: 3" in gdc_1d._format_info()

    # Test the case where the grid_size is bad
    with pytest.raises(ValueError, match="Reshaping failed! "):
        GridDC(name="bad_data", data=np.arange(5), attrs=attrs_1d)

    # Test case where the data is of too high dimensionality
    with pytest.raises(ValueError, match="Can only reshape from 1D or 2D"):
        GridDC(
            name="bad_data",
            data=np.zeros((2, 3, 4)),
            attrs=dict(grid_shape=(2, 3)),
        )

    # Test missing attribute
    with pytest.raises(ValueError, match="Missing attribute 'grid_shape'"):
        GridDC(name="missing_attr", data=np.zeros((2, 3)), attrs=dict())

    ### 2d data . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
    # Create some test data of the form (2,3,4)
    # Data in columns represents the time and data in rows the grid data
    data_2d = np.arange(2 * 3 * 4).reshape((2, 12))

    # Create a GridDC and assert that the shape is correct
    attrs_2d = dict(content="grid", grid_shape=(3, 4), index_order="F")
    gdc_2d = GridDC(name="data_2d", data=data_2d, attrs=attrs_2d)

    assert gdc_2d.shape == (2, 3, 4)
    assert "time" in gdc_2d.dims
    assert "x" in gdc_2d.dims
    assert "y" in gdc_2d.dims

    for i in range(attrs_2d["grid_shape"][0]):
        assert i == gdc_2d.data.coords["x"][i]

    for i in range(attrs_2d["grid_shape"][1]):
        assert i == gdc_2d.data.coords["y"][i]

    # The format string should contains dimension information
    assert "time: 2" in gdc_2d._format_info()
    assert "x: 3" in gdc_2d._format_info()
    assert "y: 4" in gdc_2d._format_info()

    # Assert that the data is correct and of the form
    expected_2d = np.array(
        [
            [[0, 3, 6, 9], [1, 4, 7, 10], [2, 5, 8, 11]],
            [[12, 15, 18, 21], [13, 16, 19, 22], [14, 17, 20, 23]],
        ]
    )
    assert (gdc_2d == expected_2d).all()

    ### 2d data (C-order) . . . . . . . . . . . . . . . . . . . . . . . . . . .
    data_2d_C = np.arange(2 * 3 * 4).reshape((2, 12))

    # Create a GridDC and assert that the shape is correct
    attrs_2d_C = dict(content="grid", grid_shape=(3, 4), index_order="C")
    gdc_2d_C = GridDC(name="data_2d_C", data=data_2d_C, attrs=attrs_2d_C)

    assert gdc_2d_C.shape == (2, 3, 4)
    expected_2d_C = np.array(
        [
            [[0, 1, 2, 3], [4, 5, 6, 7], [8, 9, 10, 11]],
            [[12, 13, 14, 15], [16, 17, 18, 19], [20, 21, 22, 23]],
        ]
    )
    assert (gdc_2d_C == expected_2d_C).all()


def test_griddc_integration(with_test_models):
    """Integration test for the GridDC using one of the demo models"""

    mtc = ModelTest(ADVANCED_MODEL, test_file=__file__)

    # Create and run a multiverse and load the data
    _, dm = mtc.create_run_load(
        from_cfg="../cfg/griddc_cfg.yml", parameter_space=dict(num_steps=3)
    )

    # Get the data
    grid_data = dm["multiverse"][0][f"data/{ADVANCED_MODEL}/ca"]
    print("Grid data: ", grid_data)

    # Assert the type of the state is a GridDC
    assert isinstance(grid_data, GridDC)

    # See that grid shape, extent etc. is carried over and matches
    assert "grid_shape" in grid_data.attrs
    assert "space_extent" in grid_data.attrs

    # By default, this should be a proxy
    assert grid_data.data_is_proxy
    assert "proxy" in grid_data._format_info()

    # Nevertheless, the properties should show the target values
    print("proxy:\n", grid_data._data)
    assert grid_data.ndim == 3
    assert grid_data.shape == (4,) + tuple(grid_data.attrs["grid_shape"])
    assert grid_data.grid_shape == tuple(grid_data.attrs["grid_shape"])

    # And the format string should show dimension information
    assert "time: 4" in grid_data._format_info()
    assert "x: 24" in grid_data._format_info()
    assert "y: 32" in grid_data._format_info()
    assert "cell_ids" not in grid_data._format_info()

    # ... without resolving
    assert grid_data.data_is_proxy

    # Now resolve the proxy and check the properties again
    grid_data.data
    assert not grid_data.data_is_proxy
    assert "proxy" not in grid_data._format_info()

    print("resolved:\n", grid_data._data)
    assert grid_data.ndim == 3
    assert grid_data.shape == (4,) + tuple(grid_data.attrs["grid_shape"])
    assert grid_data.grid_shape == tuple(grid_data.attrs["grid_shape"])

    assert "time: 4" in grid_data._format_info()
    assert "x: 24" in grid_data._format_info()
    assert "y: 32" in grid_data._format_info()
    assert "cell_ids" not in grid_data._format_info()

    # Check that coordinate information is available
    assert (grid_data.coords["time"] == [0, 1, 2, 3]).all()
    assert np.isclose(
        grid_data.coords["x"], np.linspace(0.0, 24.0, 24, False) + 0.5
    ).all()
    assert np.isclose(
        grid_data.coords["y"], np.linspace(0.0, 32.0, 32, False) + 0.5
    ).all()

    # Try re-instating the proxy
    grid_data.reinstate_proxy()
    assert grid_data.data_is_proxy
    assert "proxy" in grid_data._format_info()

    # And resolve it again
    grid_data.data
    assert not grid_data.data_is_proxy
    assert "proxy" not in grid_data._format_info()

    print("reinstated:\n", grid_data._data)
    assert grid_data.ndim == 3
    assert len(grid_data["time"]) == 4

    assert grid_data.shape == (4,) + tuple(grid_data.attrs["grid_shape"])
    assert grid_data.grid_shape == tuple(grid_data.attrs["grid_shape"])

    # Test some further things related to grid shape . . . . . . . . . . . . .
    cfg = dm["multiverse"][0][f"cfg/{ADVANCED_MODEL}"]
    assert grid_data.space_extent[0] == tuple(cfg["grid_shape"])[0]
    assert grid_data.space_extent[1] == tuple(cfg["grid_shape"])[1]

    # Test time coordinates again . . . . . . . . . . . . . . . . . . . . . . .
    _, dm = mtc.create_run_load(
        from_cfg="../cfg/griddc_cfg.yml",
        parameter_space=dict(num_steps=11, write_every=2),
    )
    grid_data = dm["multiverse"][0][f"data/{ADVANCED_MODEL}/ca"]
    print("Grid data: ", grid_data)
    print("Attributes: ", dict(grid_data.attrs))

    # Assert the type of the state is a GridDC and that it's loaded as proxy
    assert isinstance(grid_data, GridDC)
    assert grid_data.data_is_proxy

    # Check the time coordinates
    assert (grid_data.coords["time"] == [0, 2, 4, 6, 8, 10]).all()


@pytest.mark.skip("No hexagonal grid data available in model")
def test_griddc_integration_hexagonal():
    """Integration test for the GridDC using the ContDisease model with
    hexagonal lattice"""
    # Configure the ModelTest class for ContDisease
    mtc = ModelTest(ADVANCED_MODEL, test_file=__file__)

    # Create and run a multiverse and load the data
    update_meta_cfg = {
        "parameter_space": {
            ADVANCED_MODEL: {
                "grid_structure": "hexagonal",
                "space": {"extent": [2.217, 0.96]},
            },
        }
    }
    _, dm = mtc.create_run_load(
        from_cfg="../cfg/griddc_cfg.yml", **update_meta_cfg
    )

    # # Get the data
    grid_data = dm["multiverse"][0][f"data/{ADVANCED_MODEL}/kind"]
    grid_data.data  # resolve proxy
    print("resolved hexagonal grid data:\n", grid_data._data)

    assert "x: 24" in grid_data._format_info()
    assert "y: 32" in grid_data._format_info()

    height = 0.96 / 4 / 0.75
    size = height / 2.0
    width = 3**0.5 * size

    xs = np.linspace(0, width * 8, 8, False) + 0.5 * width
    ys = np.linspace(0, height * 0.75 * 4, 4, False) + 0.5 * height
    assert np.isclose(grid_data.coords["x"], xs, rtol=1e-03).all()
    assert np.isclose(grid_data.coords["y"], ys, rtol=1e-03).all()



"""Implements generally useful functions, partly by importing from
:py:mod:`dantro.tools`"""

import io
import logging
import os
import re
import subprocess
import sys
from datetime import timedelta
from typing import Any, Callable, Optional, Sequence, Tuple, Union

from dantro.tools import (
    IS_A_TTY,
    TTY_COLS,
    center_in_line,
    fill_line,
    format_time,
    is_iterable,
    make_columns,
    print_line,
    recursive_getitem,
    recursive_update,
)

from ._yaml import load_yml, write_yml, yaml

log = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
# working on dicts ------------------------------------------------------------


def load_selected_keys(
    src: dict,
    *,
    add_to: dict,
    keys: Sequence[Tuple[str, type, bool]],
    err_msg_prefix: str = None,
    prohibit_unexpected: bool = True,
) -> None:
    """Loads (only) selected keys from dict ``src`` into dict ``add_to``.

    Args:
        src (dict): The dict to load values from
        add_to (dict): The dict to load values into
        keys (Sequence[Tuple[str, type, bool]]): Which keys to load, given
            as sequence of ``(key, allowed types, [required=False])`` tuples.
        err_msg_prefix (str): A description string, used in error message
        prohibit_unexpected (bool, optional): Whether to raise on keys
            that were unexpected, i.e. not given in ``keys`` argument.

    Raises:
        KeyError: On missing key in ``src``
        TypeError: On bad type of value in ``src``
        ValueError: On unexpected keys in ``src``
    """

    def unpack(spec) -> Tuple[str, Union[type, Sequence[type]], bool]:
        """Unpacks a schema entry into a 3-tuple"""
        if len(spec) == 3:
            k, allowed_types, required = spec
        else:
            k, allowed_types = spec
            required = False

        return k, allowed_types, required

    for spec in keys:
        k, allowed_types, required = unpack(spec)
        if k not in src:
            if not required:
                continue
            raise ValueError(
                "{}Missing required key: {}"
                "".format(err_msg_prefix + " " if err_msg_prefix else "", k)
            )

        if not isinstance(src[k], allowed_types):
            raise TypeError(
                "{}Bad type for value of '{}'! Expected "
                "{} but got {}: {}"
                "".format(
                    err_msg_prefix + " " if err_msg_prefix else "",
                    k,
                    allowed_types,
                    type(src[k]),
                    src[k],
                )
            )

        add_to[k] = src[k]

    if not prohibit_unexpected:
        return

    allowed_keys = tuple(unpack(spec)[0] for spec in keys)
    unexpected_keys = tuple(k for k in src if k not in allowed_keys)
    if unexpected_keys:
        raise ValueError(
            "{}Received unexpected keys: {}\n"
            "Expected only: {}"
            "".format(
                err_msg_prefix + " " if err_msg_prefix else "",
                ", ".join(unexpected_keys),
                ", ".join(allowed_keys),
            )
        )


def add_item(
    value,
    *,
    add_to: dict,
    key_path: Sequence[str],
    value_func: Callable = None,
    is_valid: Callable = None,
    ErrorMsg: Callable = None,
) -> None:
    """Adds the given value to the ``add_to`` dict, traversing the given key
    path. This operation happens in-place.

    Args:
        value: The value of what is to be stored. If this is a callable, the
            result of the call is stored.
        add_to (dict): The dict to add the entry to
        key_path (Sequence[str]): The path at which to add it
        value_func (Callable, optional): If given, calls it with ``value`` as
            argument and uses the return value to add to the dict
        is_valid (Callable, optional): Used to determine whether ``value`` is
            valid or not; should take single positional argument, return bool
        ErrorMsg (Callable, optional): A raisable object that prints an error
            message; gets passed ``value`` as positional argument.

    Raises:
        Exception: type depends on specified ``ErrorMsg`` callable
    """
    # Check the value by calling the function; it should raise an error
    if is_valid is not None:
        if not is_valid(value):
            raise ErrorMsg(value)

    # Determine which keys need to be traversed
    traverse_keys, last_key = key_path[:-1], key_path[-1]

    # Set the starting point
    d = add_to

    # Traverse keys
    for _key in traverse_keys:
        # Check if a new entry is needed
        if _key not in d:
            d[_key] = dict()

        # Select the new entry
        d = d[_key]

    # Now d is where the value should be added
    # If applicable
    if value_func is not None:
        value = value_func(value)

    # Store in dict, mutable. Done.
    d[last_key] = value


# -- String formatting --------------------------------------------------------


def pprint(obj: Any, **kwargs):
    """Prints a "pretty" string representation of the given object.

    Args:
        obj (Any): The object to print
        **kwargs: Passed to print
    """
    print(pformat(obj), **kwargs)


def pformat(obj: Any) -> str:
    """Creates a "pretty" string representation of the given object.

    This is achieved by creating a yaml representation.

    .. todo::

        Improve parsing of leaf-level mappings
    """
    sstream = io.StringIO("")
    yaml.dump(obj, stream=sstream)
    sstream.seek(0)
    return sstream.read()


# misc ------------------------------------------------------------------------


def ensure_not_None(
    d: Optional[Any], fallback: Union[type, Callable] = dict
) -> Any:
    """Returns ``d`` if it is not None, otherwise creates *a new object* by
    calling ``fallback`` without any arguments."""
    if d is None:
        return fallback()
    return d


def ensure_dict(d: Optional[dict]) -> dict:
    return ensure_not_None(d, fallback=dict)


def parse_si_multiplier(s: str) -> int:
    """Parses a string like ``1.23M`` or ``-2.34 k`` into an integer.

    If it is a string, parses the SI multiplier and returns the appropriate
    integer for use as number of simulation steps.
    Supported multipliers are ``k``, ``M``, ``G`` and ``T``. These need to be
    used as the suffix of the string.

    .. note::

        This is only intended to be used with integer values and does *not*
        support float values like ``1u``.

    The used regex can be found `here <https://regex101.com/r/xngAoc/1>`_.

    Args:
        s (str): A string representing an integer number, potentially including
            a supported SI multiplier as *suffix*.

    Returns:
        int: The parsed number of steps as integer. If the value has decimal
            places, integer rounding is applied.

    Raises:
        ValueError: Upon string that does not match the expected pattern
    """
    SUFFIXES = dict(k=1e3, M=1e6, G=1e9, T=1e12)
    pattern = r"^(?P<value>\-?\s?\d+|\-?\s?\d+\.\d+)?\s?(?P<suffix>[kMGT])?$"
    # See:  https://regex101.com/r/xngAoc/1

    match = re.match(pattern, s.strip())
    if not match:
        raise ValueError(
            f"Cannot parse '{s}' into an integer! May only contain the metric "
            "suffixes k, M, G, or T. Examples: 1000, 1k, 1.23M, 0.5 G"
        )

    groups = match.groupdict()
    val = float(groups["value"].replace(" ", ""))
    mul = SUFFIXES[groups["suffix"]] if groups["suffix"] else 1

    return int(val * mul)


def parse_num_steps(
    N: Union[str, int], *, raise_if_negative: bool = True
) -> int:
    """Given a string like ``1.23M`` or an integer, prepares the num_steps
    argument for a single universe simulation.

    For string arguments, uses :py:func:`~utopya.tools.parse_si_multiplier` for
    string parsing. If that fails, attempts to read it in float notation by
    calling ``int(float(N))``.

    .. note:: This function always applies integer rounding.

    Args:
        N (Union[str, int]): The ``num_steps`` argument as a string or integer.
        raise_if_negative (bool, optional): Whether to raise an error if the
            value is negative.

    Returns:
        int: The parsed value for ``num_steps``

    Raises:
        ValueError: Result invalid, i.e. not parseable or of negative value.
    """
    if isinstance(N, str):
        try:
            N = parse_si_multiplier(N)

        except ValueError as err:
            # Don't give up just yet, could still be in scientific notation ...
            try:
                N = int(float(N))

            except:
                # Ok, that also failed. Giving up now.
                raise ValueError(f"Failed parsing `num_steps`: {err}") from err

    if raise_if_negative and N < 0:
        raise ValueError(
            f"Argument `num_steps` needs to be non-negative, but was {N}!"
        )

    return N



"""Provides classes that can be used for model implementation:

- :py:class:`~utopya_backend.model.base.BaseModel`: provides shared simulation
  infrastructure like a logger, a shared RNG and signal handling.
- :py:class:`~utopya_backend.model.step.StepwiseModel`: a base model class
  that is optimized for stepwise model iteration.

All these base models still require to be subclassed and certain methods being
implemented.
"""

# isort: skip_file

from .base import BaseModel
from .step import StepwiseModel



"""Implements the utopya run-existing CLI subtree"""

import os

import click

from ._shared import (
    OPTIONS,
    add_options,
    complete_model_names,
    complete_run_dirs,
)


@click.command(
    help=(
        "[EXPERIMENTAL] (Re-)Run universes of an existing simulation run.\n"
        "\n"
        "Restores a run of MODEL_NAME from the given RUN_DIR. "
        "Subsequently, individual universes can be (re-)run.\n"
        "\n"
        "Note that this feature is currently experimental, meaning that the "
        "interface may still change a lot."
    ),
)
@click.argument("model_name", shell_complete=complete_model_names)
@click.argument(
    "run_dir",
    shell_complete=complete_run_dirs,
    required=True,
)
@add_options(OPTIONS["label"])
@click.option(
    "-u",
    "--uni",
    "--universe",
    "universes",
    multiple=True,
    type=str,
    help=(
        "Which universes to run (e.g.: 00154). Note that leading zeros need "
        "to be added. To supply multiple, use the -u option multiple times. "
        "If no universes are specified, a run on all universes is performed."
    ),
)
@click.option(
    "-c",
    "--clear-existing",
    "clear_existing_output",
    default=False,
    is_flag=True,
    help=(
        "Whether to clear existing output files from universes. "
        "Set this option to re-run universes that were previously run. "
        "Cannot be used together with --skip-existing."
    ),
)
@click.option(
    "-s",
    "--skip-existing",
    "skip_existing_output",
    default=False,
    is_flag=True,
    help=(
        "Whether to skip universes with existing output. "
        "Set this option to complete universes from a previous run. "
        "Cannot be used together with --uni or --clear-existing."
    ),
)
@add_options(OPTIONS["num_workers"])  # -W, --num-workers
#
#
#
@click.pass_context
def run_existing(
    ctx,
    run_dir,
    model_name: str,
    label: str,
    universes: list,
    num_workers: int,
    clear_existing_output: bool,
    skip_existing_output: bool,
    **kwargs,
):
    """Repeats a model simulation in parts or entirely"""
    import utopya

    _log = utopya._getLogger("utopya")

    model = utopya.Model(name=model_name, bundle_label=label)
    mv = model.create_distributed_mv(run_dir=run_dir)

    if universes:
        if skip_existing_output:
            raise RuntimeError(
                "Option --skip-existing cannot be set together "
                "with a list of universes to perform."
            )

        mv.run_selection(
            uni_id_strs=universes,
            num_workers=num_workers,
            clear_existing_output=clear_existing_output,
        )
    else:
        if skip_existing_output and clear_existing_output:
            raise RuntimeError(
                "Options --skip-existing and --clear-existing are exclusive "
                "but both were set."
            )

        mv.run(
            num_workers=num_workers,
            clear_existing_output=clear_existing_output,
            skip_existing_output=skip_existing_output,
        )

    _log.note("Evaluation routine is not possible for repeated run.")
    _log.remark(
        "For evaluation, call:\n  utopya eval %s %s\n",
        model_name,
        os.path.basename(run_dir),
    )
    _log.progress("Exiting now ...\n")



"""Test module for example plot generation"""

import contextlib
import os

import pytest

from utopya.eval import PlotHelper, PlotManager
from utopya.testtools import ModelTest
from utopya.tools import load_yml

from .. import *
from .._fixtures import *
from ..eval.test_plotting import (
    ABM_PLOTS_CFG,
    HEXGRID_PLOTS_CFG,
    ObjectContainer,
    XarrayDC,
    abm_data,
    hexgrid_data,
)

PLOTS_CFG = os.path.join(os.path.dirname(__file__), "test_plots.yml")
"""Path to the plots configuration file"""

# -----------------------------------------------------------------------------


@pytest.fixture(autouse=True)
def register_demo_project(tmp_projects, with_test_models):
    """Use on all tests in this module"""
    pass


# -----------------------------------------------------------------------------


def test_plots(dm, out_dir):
    """Creates output from the (DAG-based) plotting tests and examples"""

    model = ModelTest(ADVANCED_MODEL)
    mv, dm = model.create_run_load(parameter_space=dict(num_steps=42))
    mv.pm.raise_exc = True
    print(dm.tree)

    # Load some configuration arguments
    shared_kwargs = dict(out_dir=str(out_dir))
    plot_cfgs = load_yml(PLOTS_CFG)

    # Can do a simple DAG-based universe and multiverse plot
    for cfg_name, plot_cfg in plot_cfgs.items():
        if cfg_name.startswith("."):
            continue

        _raises = plot_cfg.pop("_raises", None)
        _match = plot_cfg.pop("_match", None)

        if _raises is not None:
            ctx = pytest.raises(globals()[_raises], match=_match)
        else:
            ctx = contextlib.nullcontext()

        # The actual plotting
        print(f"\n\n--- Test case '{cfg_name}' ---")
        with ctx:
            mv.pm.plot(cfg_name, **shared_kwargs, **plot_cfg)


def test_imshow_hex(hexgrid_data, out_dir):
    """Creates output that illustrates imshow_hexagonal"""

    from utopya.eval.plots.ca import imshow_hexagonal as imshow_hex

    def new_helper(name: str, *, ext: str = "pdf", **kws):
        out_path = os.path.join(out_dir, f"{name}.{ext}")
        os.makedirs(os.path.dirname(out_path), exist_ok=True)
        ph = PlotHelper(out_path=out_path, **kws)
        ph.setup_figure()
        return ph

    def draw_imshow_hex(name: str, data, **kwargs):
        hlpr = new_helper(name)
        imshow_hex(data, ax=hlpr.ax, **kwargs)
        hlpr.save_figure()

    # .........................................................................
    data_small = hexgrid_data["small"].sel(time=0, drop=True)
    space_kwargs = dict(space_size=(8, 8), space_offset=(-4, -4))

    draw_imshow_hex(
        "small",
        data_small,
    )

    draw_imshow_hex(
        "small_centers_marked",
        data_small,
        draw_centers=True,
    )

    draw_imshow_hex(
        "small_with_space_outer",
        data_small,
        update_grid_properties=dict(**space_kwargs),
    )

    draw_imshow_hex(
        "small_with_space_inner",
        data_small,
        update_grid_properties=dict(
            **space_kwargs,
            space_boundary="inner",
        ),
    )

    draw_imshow_hex(
        "small_flat_top_odd",
        data_small,
        update_grid_properties=dict(pointy_top=False, offset_mode="odd"),
    )

    draw_imshow_hex(
        "small_flat_top_odd_with_space",
        data_small,
        update_grid_properties=dict(
            pointy_top=False, offset_mode="odd", **space_kwargs
        ),
    )


def test_caplot(hexgrid_data, out_dir):
    """Creates output that illustrates imshow_hexagonal"""
    # Add the hexgrid data to the data tree
    mv, _ = ModelTest(ADVANCED_MODEL).create_run_load()

    for _, uni in mv.dm["multiverse"].items():
        hexdata = uni[("data", ADVANCED_MODEL)].new_group("hexgrid")
        for name, data in hexgrid_data.items():
            hexdata.new_container(name, data=data, Cls=XarrayDC)

    # Configure PlotManager
    mv.pm.raise_exc = True
    mv.pm._out_dir = out_dir

    # Plot only the documentation-related plots
    for name, plot_cfg in load_yml(HEXGRID_PLOTS_CFG).items():
        if not name.startswith("doc_"):
            continue

        try:
            mv.pm.plot(name[4:], **plot_cfg)

        except Exception as exc:
            # Allow to fail if it's due to ffmpeg missing
            if "ffmpeg" in str(exc) and HAVE_FFMPEG:
                raise
            continue


def test_abmplot(abm_data, out_dir):
    """Creates output that illustrates abmplot"""
    # Add the data to the data tree
    mv, _ = ModelTest(ADVANCED_MODEL).create_run_load()

    for _, uni in mv.dm["multiverse"].items():
        grp = uni[("data", ADVANCED_MODEL)].new_group("abm")
        for name, data in abm_data.items():
            grp.new_container(name, data=data, Cls=ObjectContainer)

    mv.pm.raise_exc = True
    mv.pm._out_dir = out_dir

    # Plot only the documentation-related plots
    for name, plot_cfg in load_yml(ABM_PLOTS_CFG).items():
        if not name.startswith("doc_"):
            continue

        try:
            mv.pm.plot(name[4:], **plot_cfg)

        except Exception as exc:
            # Allow to fail if it's due to ffmpeg missing
            if "ffmpeg" in str(exc) and HAVE_FFMPEG:
                raise
            continue



"""Test the plot_funcs._utils module"""

import itertools

import numpy as np
import pytest
import xarray as xr
from pkg_resources import resource_filename

import utopya
import utopya.eval.plots._utils

# Fixtures --------------------------------------------------------------------


# Tests -----------------------------------------------------------------------


def test_calc_pxmap_rectangles():
    """Test the calc_pxmap_rectangles function"""

    cpr = utopya.eval.plots._utils.calc_pxmap_rectangles

    def create_2D_data(
        *shape, dims=None, coords_func=np.linspace, rg_args=(0.0, 1.0)
    ):
        data = xr.DataArray(np.random.random(shape), dims=dims)
        data.coords[data.dims[0]] = coords_func(*rg_args, shape[0])
        data.coords[data.dims[1]] = coords_func(*rg_args, shape[1])
        return data

    # Basics . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
    shape = (6, 11)
    data = create_2D_data(*shape, dims=("foo", "bar"))
    ds, lims = cpr(
        x_coords=data.coords["foo"].data, y_coords=data.coords["bar"].data
    )

    # Limits should be [0., 1.] for both dimensions
    assert lims["x"] == (0.0, 1.0)
    assert lims["y"] == (0.0, 1.0)

    # The dataset should contain rectangle information for each point
    assert ds["pos_x"].shape == shape
    assert ds["pos_y"].shape == shape
    assert ds["len_x"].shape == shape
    assert ds["len_y"].shape == shape

    # Check the shape of some exemplary rectangles
    pt = ds.sel(x=0.0, y=0.0)
    assert pt["pos_x"] == 0.0
    assert pt["pos_y"] == 0.0
    assert pt["len_x"] == 0.1
    assert pt["len_y"] == 0.05

    # With extended rectangles, the limits are adjusted . . . . . . . . . . . .
    shape = (4, 2)
    data = create_2D_data(*shape)
    ds, lims = cpr(
        x_coords=data.coords["dim_0"].data,
        y_coords=data.coords["dim_1"].data,
        extend=True,
    )

    ext_x, ext_y = 0.5 / (shape[0] - 1), 0.5 / (shape[1] - 1)
    assert lims["x"] == (0.0 - ext_x, 1.0 + ext_x)
    assert lims["y"] == (0.0 - ext_y, 1.0 + ext_y)

    # The corresponding point should be extended beyond the coordinate space
    pt = ds.sel(x=0.0, y=0.0)
    print(pt)
    assert pt["pos_x"] == 0.0 - ext_x
    assert pt["pos_y"] == 0.0 - ext_y
    assert pt["len_x"] == 1.0 / (shape[0] - 1)
    assert pt["len_y"] == 1.0 / (shape[1] - 1)
    assert (pt["rect_spec"].values["xy"] == [pt["pos_x"], pt["pos_y"]]).all()
    assert pt["rect_spec"].values["width"] == pt["len_x"]
    assert pt["rect_spec"].values["height"] == pt["len_y"]

    # Unpacking also works; need .item call though
    (lambda *args: args)(*pt["rect_spec"].item())

    # Should be possible to only give one coordinate . . . . . . . . . . . . .
    ds, lims = cpr(
        x_coords=data.coords["dim_0"].data, y_coords=None, default_distance=3.0
    )

    assert lims["y"] == (-3 / 2, +3 / 2)

    ds, lims = cpr(
        x_coords=data.coords["dim_0"].data, y_coords=None, extend=True
    )
    assert lims["y"] == (-0.5, +0.5)

    # There is full coverage of the space . . . . . . . . . . . . . . . . . . .
    shapes = [(5, 11), (2, 3), (1, 2), (2, 1), (1, 1)]

    for ax, shape, ext in itertools.product(("x", "y"), shapes, (True, False)):
        print(
            "\n --- Checking coverage in {}-direction for shape {}, {} ---"
            "".format(ax, shape, "extended" if ext else "not extended")
        )

        # Create data
        data = create_2D_data(*shape)
        ds, lims = cpr(
            x_coords=data.coords["dim_0"].data,
            y_coords=data.coords["dim_1"].data,
            extend=ext,
        )
        print("Limits:", lims[ax])

        # Determine lower borders
        b_lower = ds["pos_" + ax]
        print("Lower borders:", b_lower)

        # Determine upper borders
        b_upper = ds["pos_" + ax] + ds["len_" + ax]
        print("Upper borders:", b_upper)

        # Upper border of one rectangle should coincide with lower of the next
        if b_lower.sizes[ax] > 1 and b_upper.sizes[ax] > 1:
            assert np.isclose(
                b_upper[{ax: slice(None, -1)}].values,
                b_lower[{ax: slice(1, None)}].values,
            ).all()
        # else: only a single rectangle in this direction

        # First rectangle's lower border is the lower limit
        assert np.isclose(b_lower[{ax: 0}].values, lims[ax][0]).all()

        # Last rectangle's upper border is the upper limit
        assert np.isclose(b_upper[{ax: -1}].values, lims[ax][1]).all()

    # Change scales . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
    # The ones below are only coverage tests

    shape = (5, 9)
    data = create_2D_data(*shape, coords_func=np.logspace, rg_args=(2, 6))

    # Logarithmic
    _, _ = cpr(
        x_coords=data.coords["dim_0"].data,
        y_coords=data.coords["dim_1"].data,
        x_scale="lin",
        y_scale="log",
    )

    _, _ = cpr(
        x_coords=data.coords["dim_0"].data,
        y_coords=data.coords["dim_1"].data,
        x_scale="log",
        y_scale="lin",
    )

    _, _ = cpr(
        x_coords=data.coords["dim_0"].data,
        y_coords=data.coords["dim_1"].data,
        x_scale="log",
        y_scale="log",
    )

    # Categorical
    _, _ = cpr(
        x_coords=data.coords["dim_0"].data,
        y_coords=data.coords["dim_1"].data,
        x_scale="cat_at_value",
        y_scale="categorical",
    )

    # Make sure the values are as expected
    ds, lims = cpr(
        x_coords=data.coords["dim_0"].data,
        y_coords=data.coords["dim_1"].data,
        x_scale="log",
        y_scale="log",
        extend=True,
    )

    assert np.isclose(lims["x"], [10**1.5, 10**6.5], rtol=1.0e-6).all()
    assert np.isclose(lims["y"], [10**1.75, 10**6.25], rtol=1.0e-6).all()

    pt = ds.sel(x=1.0e2, y=1.0e2)
    assert np.isclose(pt["pos_x"], 10**1.5)
    assert np.isclose(pt["pos_y"], 10**1.75)
    assert np.isclose(pt["len_x"], 10**2.5 - 10**1.5)
    assert np.isclose(pt["len_y"], 10**2.25 - 10**1.75)

    # Errors . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
    with pytest.raises(ValueError, match="Invalid scale argument"):
        cpr(
            x_coords=data.coords["dim_0"].data,
            y_coords=data.coords["dim_1"].data,
            x_scale="foo",
        )

    with pytest.raises(ValueError, match="need to be at least of length one"):
        cpr(x_coords=[], y_coords=None)



"""A module that implements python functions that generate figures which
are embedded into the documentation.
These are implemented as pytests to allow using the fixtures defined throughout
the utopya test suite.

NOTE Test function names should ideally be unique within this subpackage!

     This is because this test module is run from within the sphinx build
     routine (see doc/conf.py) and sets ``UTOPYA_ABBREVIATE_TEST_OUTPUT_DIR``,
     thus potentially allowing to overwrite files from other tests if the
     test functions have the same name.
"""



"""Implements a plotting framework based on dantro

In order to make the plotting framework specific to Utopia, this module derives
both from the dantro PlotManager and some PlotCreator classes.
"""

import contextlib
import copy
import importlib
import logging
import os
import sys
from typing import Dict, Union

import dantro
import dantro._import_tools
import dantro.plot.creators
import dantro.plot.utils
import dantro.plot_mngr

from ..model_registry import ModelInfoBundle
from ._plot_func_resolver import PlotFuncResolver
from .plotcreators import (
    MultiversePlotCreator,
    PyPlotCreator,
    UniversePlotCreator,
)

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


class PlotManager(dantro.plot_mngr.PlotManager):
    """This is the Utopia-specific version of the dantro ``PlotManager``.

    It registers the Utopia-specific plot creators and allows for custom
    interface specifications, e.g. by preloading custom modules.
    """

    CREATORS: Dict[str, type] = dict(
        base=dantro.plot.creators.BasePlotCreator,
        external=PyPlotCreator,
        pyplot=PyPlotCreator,
        universe=UniversePlotCreator,
        multiverse=MultiversePlotCreator,
    )
    """Supported plot creator classes"""

    MODEL_PLOTS_MODULE_NAME = "model_plots"
    """Name under which the model-specific plots are made importable"""

    PLOT_FUNC_RESOLVER: type = PlotFuncResolver
    """The custom plot function resolver type to use."""

    # .........................................................................

    def __init__(
        self, *args, _model_info_bundle: ModelInfoBundle = None, **kwargs
    ):
        """Sets up a PlotManager.

        This specialization of the :py:class:`dantro.plot_mngr.PlotManager`
        additionally stores some utopya-specific metadata in form of a
        :py:class:`~utopya.model_registry.info_bundle.ModelInfoBundle` that
        describes the model this PlotManager is used with.
        That information is then used to load some additional model-specific
        information once a creator is invoked.

        Furthermore, the :py:meth:`._preload_modules` method takes care to make
        model-, project-, or framework-specific plot functions available.

        Args:
            *args: Positional arguments passed to
                :py:class:`~dantro.plot_mngr.PlotManager`.
            _model_info_bundle (ModelInfoBundle, optional): The internally-used
                argument to pass model information to the plot manager.
            **kwargs: Keyword arguments passed on to
                :py:class:`~dantro.plot_mngr.PlotManager`.
        """
        super().__init__(*args, **kwargs)

        self._model_info_bundle = copy.deepcopy(_model_info_bundle)

        self._preload_modules()

    @property
    def common_out_dir(self) -> str:
        """The common output directory of all plots that were created with
        this plot manager instance. This uses the plot output paths stored in
        the plot information dict, specifically the ``target_dir`` entry.

        If there was no plot information yet, the return value will be empty.
        """
        p = os.path.commonprefix([d["target_dir"] for d in self.plot_info])
        if not os.path.exists(p):
            p = os.path.dirname(p)
        return p

    def plot_from_cfg(
        self, *args, plots_cfg: Union[str, dict] = None, **kwargs
    ):
        """Thin wrapper around parent method that shows which plot
        configuration file will be used.
        """
        log.hilight(
            "Now creating plots for '%s' model ...",
            self._model_info_bundle.model_name,
        )
        if isinstance(plots_cfg, str):
            log.note("Plots configuration:\n  %s\n", plots_cfg)

        elif plots_cfg is None:
            log.note(
                "Using default plots configuration:\n  %s\n",
                self._model_info_bundle.paths.get("default_plots"),
            )

        return super().plot_from_cfg(*args, plots_cfg=plots_cfg, **kwargs)

    def _get_plot_func_resolver(self, **init_kwargs) -> PlotFuncResolver:
        """Instantiates the plot function resolver object.

        Additionally attaches the model info bundle to the resolver, such that
        it can use that information for plot function lookup.
        """
        return self.PLOT_FUNC_RESOLVER(
            **init_kwargs, _model_info_bundle=self._model_info_bundle
        )

    def _get_plot_creator(
        self, *args, **kwargs
    ) -> dantro.plot.creators.BasePlotCreator:
        """Sets up the BasePlotCreator and attaches a model information bundle
        to it such that this information is available downstream.
        """
        creator = super()._get_plot_creator(*args, **kwargs)
        creator._model_info_bundle = copy.deepcopy(self._model_info_bundle)

        return creator

    def _preload_modules(self):
        """Pre-loads the model-, project-, and framework-specific plot
        function modules.
        This allows to execute code (like registering model-specific dantro
        data operations) and have them available prior to the invocation of
        the creator and independently from the module that contains the plot
        function (which may be part of dantro, for instance).

        Uses :py:func:`dantro._import_tools.import_module_from_path`
        """
        import_module_from_path = dantro._import_tools.import_module_from_path
        # . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
        # A simple exception handling context

        @contextlib.contextmanager
        def exception_handling(ExcType, scope: str):
            try:
                yield

            except ExcType as exc:
                _msg = (
                    f"{scope.title()}-specific plot module "
                    "could not be imported!"
                )
                if self.raise_exc:
                    raise ExcType(
                        f"{_msg}\n\nError was: {exc}\n\n"
                        "For debugging, inspect the traceback. Disable debug "
                        "mode to ignore exception and continue, even if this "
                        "may cause errors during plotting."
                    ) from exc
                log.warning(_msg)
                log.caution(
                    "This may lead to errors during plotting. "
                    "Enable debug mode to get a traceback."
                )

        # . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

        mib = self._model_info_bundle
        _preloaded = []

        if mib is not None:
            log.note("Pre-loading plot modules ...")

            mod_path = mib.paths.get("py_plots_dir")
            if mod_path and os.path.exists(mod_path):
                with exception_handling(ImportError, "model"):
                    log.debug("  Loading model-specific plot module ...")
                    _ms = f"{self.MODEL_PLOTS_MODULE_NAME}.{mib.model_name}"
                    import_module_from_path(
                        mod_path=mod_path,
                        mod_str=_ms,
                    )
                    _preloaded.append("model")

            # Also do this on the project and framework level
            # TODO Should make module name configurable separately! See #9
            project = mib.project
            if (
                project
                and project.paths.py_plots_dir
                and project.settings.preload_project_py_plots in (None, True)
            ):
                with exception_handling(ImportError, "project"):
                    log.debug("  Loading project-specific plot module ...")
                    dantro._import_tools.import_module_from_path(
                        mod_path=project.paths.py_plots_dir,
                        mod_str=f"{self.MODEL_PLOTS_MODULE_NAME}",
                    )
                    _preloaded.append("project")

            if (
                project
                and project.framework_project
                and project.settings.preload_framework_py_plots in (None, True)
            ):
                fw = project.framework_project
                if fw and fw.paths.py_plots_dir:
                    with exception_handling(ImportError, "framework"):
                        log.debug(
                            "  Loading framework-specific plot module ..."
                        )
                        import_module_from_path(
                            mod_path=fw.paths.py_plots_dir,
                            mod_str=f"{self.MODEL_PLOTS_MODULE_NAME}",
                        )
                        _preloaded.append("framework")

        if _preloaded:
            log.remark(
                "  Pre-loaded plot modules of:  %s", ", ".join(_preloaded)
            )
        else:
            log.remark(
                "  No `py_plots_dir` available or pre-loading deactivated."
            )



"""Defines arguments that are shared across various parts of the CLI"""

import glob
import os
from typing import Dict, List, Union

import click

# -----------------------------------------------------------------------------
# Variables
#
# NOTE Some of these are duplicated rather than imported to save on utopya
#      import time...

UTOPYA_CFG_DIR: str = os.path.expanduser("~/.config/utopya")
"""Directory where utopya stores its metadata"""

UTOPYA_CFG_FILE_NAMES = dict(
    user="user_cfg.yml",
    utopya="utopya_cfg.yml",
    batch="batch_cfg.yml",
)
"""Names and paths of valid configuration entries"""

UTOPYA_CFG_FILE_PATHS = {
    k: os.path.join(UTOPYA_CFG_DIR, fname)
    for k, fname in UTOPYA_CFG_FILE_NAMES.items()
}
"""Absolute configuration file paths"""

UTOPYA_CFG_SUBDIR_NAMES = dict(
    models="models",
    projects="projects",
)
"""Names and paths of valid configuration subdirectories"""

UTOPYA_CFG_SUBDIRS = {
    k: os.path.join(UTOPYA_CFG_DIR, dirname)
    for k, dirname in UTOPYA_CFG_SUBDIR_NAMES.items()
}
"""Absolute configuration file paths"""

DEFAULT_RUN_DIR_SEARCH_PATHS: str = [
    "~/utopya_output",
    "~/utopia_output",
]
"""Default directory paths to search for model run directories in.

This can be overwritten via the utopya package configuration file and its
entry ``cli.run_dir_search_paths``.
"""


# .............................................................................

INTERACTIVE_MODE_PROHIBITED_ARGS = (
    "run_cfg",
    "run_dir",
    "label",
    "set_params",
    "cluster_mode",
    "show_data_tree",
    "use_data_tree_cache",
    "load_parallel",
)
"""Argument names that may NOT be given in the interactive plotting prompt"""


# -----------------------------------------------------------------------------
# Shell completion


def complete_from_cfg_dir(
    ctx, param, incomplete: str, *, dirpath: str, glob_str: str = "*.yml"
) -> List[str]:
    """Reads the filenames from a directory and uses that to return a list of
    strings that offer

    This is meant for completing queries where a name is required that has an
    equivalent representation as a registry file in a utopya config directory.
    """
    return sorted(
        [
            os.path.splitext(os.path.basename(f))[0]
            for f in glob.glob(os.path.join(dirpath, incomplete + glob_str))
        ],
        key=str.casefold,
    )


def complete_model_names(*args) -> List[str]:
    """Completes model names using :py:func:`.complete_from_cfg_dir`."""
    return complete_from_cfg_dir(*args, dirpath=UTOPYA_CFG_SUBDIRS["models"])


def complete_project_names(*args) -> List[str]:
    """Completes project names using :py:func:`.complete_from_cfg_dir`."""
    return complete_from_cfg_dir(*args, dirpath=UTOPYA_CFG_SUBDIRS["projects"])


def complete_run_dirs(
    ctx, param, incomplete: str, *, extra_search_dirs: list = []
) -> List[str]:
    """Completes run directories for the selected model name.

    As the run directory is determined via the run configuration, there is no
    certainty on the location of run directories. Instead, the canonical
    locations where the simulation output is stored forms the basis for the
    completion.
    The search directories can be configured in the utopya package config file
    using the ``cli.run_dir_search_paths`` key:

    .. code-block:: yaml

        # ~/.config/utopya/utopya_cfg.yml
        ---
        cli:
          run_dir_search_paths:
            - ~/utopya_output
            - ~/utopia_output
            # ... can add more here ...

    .. todo::

        Auto-complete local paths as well, starting from CWD.
    """
    from dantro.tools import load_yml

    # Need the model name
    model_name = ctx.params["model_name"]

    # Get search directories from config and assemble model output directories
    cli_cfg = {}
    if os.path.exists(UTOPYA_CFG_FILE_PATHS["utopya"]):
        cli_cfg = load_yml(UTOPYA_CFG_FILE_PATHS["utopya"]).get("cli", {})

    search_dirs = cli_cfg.get(
        "run_dir_search_paths", DEFAULT_RUN_DIR_SEARCH_PATHS
    )
    search_dirs += extra_search_dirs
    model_out_dirs = [
        os.path.join(os.path.expanduser(d), model_name) for d in search_dirs
    ]

    # Aggregate candidate directory names, then sort and filter them
    candidates = []
    for model_out_dir in model_out_dirs:
        if not os.path.isdir(model_out_dir):
            continue

        candidates += [
            os.path.basename(p)
            for p in glob.glob(os.path.join(model_out_dir, "*"))
            if os.path.isdir(p)
        ]

    # TODO Fall back to auto-completion of local paths, if possible

    return [
        p for p in reversed(sorted(candidates)) if p.startswith(incomplete)
    ]


# -----------------------------------------------------------------------------
# Shared options


def add_options(options):
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options


def default_none(ctx, _, value) -> Union[None, tuple]:
    """For use in ``multiple=True`` with ``nargs=-1`` where None is desired
    for a default value instead of an empty tuple.
    Pass this to the ``callback`` argument of an argument or an option.
    """
    if len(value) == 0:
        return None
    return value


# .............................................................................

OPTIONS = dict()

# -- Model selection
OPTIONS["label"] = (
    click.option(
        "--label",
        default=None,
        help=(
            "For model names that have multiple info bundles registered, a "
            "label is needed to unambiguously select the desired one. "
            "Alternatively, use the ``utopya models set-default`` CLI command "
            "to set a default label for a model."
        ),
    ),
)

OPTIONS["model_selection"] = (
    click.option(
        "--cfg-set",
        "--cs",
        default=None,
        help=(
            "If the chosen model provides configuration sets, use the config "
            "files from the chosen set for the run and plots config. "
            "Note that the specific arguments still take precedence over the "
            "values from the config sets; to use default paths, "
            "specify empty strings (``''``) for those arguments."
        ),
    ),
) + OPTIONS["label"]

# -- Universal flags
OPTIONS["debug_flag"] = (
    click.option(
        "-d",
        "--debug",
        count=True,
        help=(
            "The debug level."
            # TODO Expand
        ),
    ),
)
OPTIONS["cluster_mode"] = (
    click.option(
        "--cluster-mode",
        flag_value=True,
        default=None,
        help="Enables cluster mode.",
    ),
)
OPTIONS["num_workers"] = (
    click.option(
        "-W",
        "--num-workers",
        default=None,
        type=click.IntRange(min=-os.cpu_count() + 1, max=+os.cpu_count()),
        help=(
            "Shortcut for meta-config entry ``worker_manager.num_workers``, "
            "which sets the number of worker processes. "
            "Can be an integer; if negative, will deduce the number from the "
            "number of available CPUs."
        ),
    ),
)

# -- Data loading options
OPTIONS["load"] = (
    click.option(
        "-P",
        "--load-parallel",
        flag_value=True,
        default=None,
        help=("If given, will force loading data in parallel."),
    ),
    click.option(
        "--use-data-tree-cache",
        "--tc",
        flag_value=True,
        default=None,
        help=(
            "If set, uses tree file caching: If no cache file exists, creates "
            "one after loading all data; if a tree file already exists, uses "
            "that to restore the data tree. This may bring a speed-up if the "
            "creation of the data tree takes a long time."
        ),
    ),
    click.option(
        "--show-data-tree",
        type=click.Choice(("full", "condensed", "none")),
        default="condensed",
        show_default=True,
        help="Controls which kind of data tree should be shown after loading.",
    ),
)


# -- Evaluation options
OPTIONS["eval"] = (
    click.option(
        "--plots-cfg",
        default=None,
        type=click.Path(exists=True, dir_okay=False, resolve_path=True),
        help=(
            "If given, uses the plots configuration file found at this path "
            "instead of the defaults for the model."
        ),
    ),
    click.option(
        "-u",
        "--update-plots-cfg",
        multiple=True,
        callback=default_none,
        help=(
            "Sets entries in the selected plots config. "
            "Example: ``-u my_plot.some_param=42`` sets the ``some_param`` "
            "entry in the plot configuration named ``my_plot``. "
            "Repeat the ``-u`` option to set multiple values."
        ),
    ),
    click.option(
        "--plot-only",
        "--po",
        "plot_only",
        multiple=True,
        callback=default_none,
        help=(
            "If given, will plot only those entries of the plot configuration "
            "that match the names given here. This can also be used to "
            "activate plots that are disabled in the specified plot "
            "configuration. Note that simple name globbing is supported, but "
            "the argument needs to be put into quotes to not conflict with "
            "the globbing done by the shell. "
            "Repeat the ``--po`` option to denote multiple ``plot_only`` "
            "arguments."
        ),
    ),
    click.option(
        "-i",
        "--interactive",
        flag_value=True,
        help=(
            "If set, the CLI will not exit after plotting finished, but allow "
            "to continue plotting in an interactive session. "
            "This option is useful for creating multiple plots in an "
            "iterative fashion, especially if data loading time is large. "
            "Note that all Multiverse-related configuration options can not "
            "be changed during the session and have to be set beforehand."
        ),
    ),
    click.option(
        "-R",
        "--reveal-output",
        flag_value=True,
        help="If set, opens the output directory after plotting finished.",
    ),
)



"""Various utilities used within the CLI definition and for handling click"""

import copy
import logging
import os
import sys
from typing import Tuple

import click
import paramspace as psp

log = logging.getLogger(__name__)
# FIXME This logger does not support all levels that are used throughout this
#       module, because it is setup prior to the utopya import.

# -----------------------------------------------------------------------------
# Communication via Terminal
# TODO Consider mapping directly to logger?


class ANSIesc:
    """Some selected ANSI escape codes; usable in format strings"""

    RESET = "\033[0m"
    BOLD = "\033[1m"
    DIM = "\033[2m"
    UNDERLINE = "\033[4m"

    BLACK = "\033[30m"
    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"
    WHITE = "\033[37m"

    ORANGE = "\033[38;5;202m"


def _parse_msg(msg: str, args) -> str:
    if args:
        return msg % args
    return msg


class Echo:
    """Adds some standardized styled ``click.echo`` calls.

    The styles are aligned with those set in the utopya.logging module.
    """

    @staticmethod
    def help(*, exit: int = None):
        """Shows the help message of the current context"""
        click.echo(click.get_current_context().get_help())
        if exit is not None:
            sys.exit(exit)

    @staticmethod
    def trace(msg: str, *args, dim=True, **style):
        """An echo that communicates some debug-level information"""
        click.secho(_parse_msg(msg, args), dim=dim, **style)

    @staticmethod
    def debug(msg: str, *args, dim=True, **style):
        """An echo that communicates some debug-level information"""
        click.secho(_parse_msg(msg, args), dim=dim, **style)

    @staticmethod
    def remark(msg: str, *args, fg=246, **style):
        """An echo that communicates some low-level information"""
        click.secho(_parse_msg(msg, args), fg=fg, **style)

    @staticmethod
    def note(msg: str, *args, fg="cyan", **style):
        """An echo that communicates some low-level information"""
        click.secho(_parse_msg(msg, args), fg=fg, **style)

    @staticmethod
    def info(msg: str, *args, **style):
        """An echo that communicates some information"""
        click.secho(_parse_msg(msg, args), **style)

    @staticmethod
    def progress(msg: str, *args, fg="green", **style):
        """An echo that communicates some progress"""
        click.secho(_parse_msg(msg, args), fg=fg, **style)

    @staticmethod
    def caution(msg: str, *args, fg=202, **style):
        """An echo that communicates a cautioning message"""
        click.secho(_parse_msg(msg, args), fg=fg, **style)

    @staticmethod
    def hilight(msg: str, *args, fg="yellow", bold=True, **style):
        """An echo that highlights a certain"""
        click.secho(_parse_msg(msg, args), fg=fg, bold=bold, **style)

    @staticmethod
    def success(msg: str, *args, fg="green", bold=True, **style):
        """An echo that communicates a success"""
        click.secho(_parse_msg(msg, args), fg=fg, bold=bold, **style)

    @staticmethod
    def warning(msg: str, *args, fg=202, bold=True, **style):
        """An echo that communicates a warning"""
        click.secho(_parse_msg(msg, args), fg=fg, bold=bold, **style)

    @staticmethod
    def error(
        msg: str,
        *args,
        error: Exception = None,
        fg="red",
        bold=True,
        **style,
    ):
        """An echo that can be used to communicate an error, optionally
        parsing the exception's error msg as well.
        """
        click.secho(_parse_msg(msg, args), fg=fg, bold=bold, **style)
        if not error:
            return

        click.secho(
            f"{type(error).__name__}: {error}", fg=fg, bold=False, **style
        )


# -----------------------------------------------------------------------------
# Parsing of key-value pairs


class _DEL:
    """Objects of this class symbolize deletion"""

    def __str__(self) -> str:
        return "<DELETION MARKER>"


DEL = _DEL()
"""An object denoting deletion"""


def convert_value(
    val: str,
    *,
    allow_deletion: bool = True,
    allow_yaml: bool = False,
    allow_eval: bool = False,
):
    """Attempts a number of conversions for a given string

    .. todo::

        Should this be here or is there a simpler way to do this?

    Args:
        val (str): Description
        allow_deletion (bool, optional): If set, can pass a ``DELETE`` string
            to a key to remove the corresponding entry.
        allow_yaml (bool, optional): Whether to attempt converting values
            by employing a YAML parser
        allow_eval (bool, optional): Whether to try calling eval() on the
            value strings during conversion
    """
    # Boolean
    if val.lower() in ("true", "false"):
        return bool(val.lower() == "true")

    # None
    if val.lower() == "null":
        return None

    # Numbers
    try:
        return int(val)
    except:
        try:
            return float(val)
        except:
            pass

    # Deletion placeholder
    if val == "DELETE":
        return DEL

    # YAML
    if allow_yaml:
        from utopya.yaml import yaml

        try:
            return yaml.load(val)
        except:
            pass

    # Last resort, if allowed: eval
    if allow_eval:
        try:
            return eval(val)
        except:
            pass

    # Just return the string
    return val


def set_entries_from_kv_pairs(
    *pairs,
    add_to: dict,
    _log=log,
    attempt_conversion: bool = True,
    **conversion_kwargs,
) -> None:
    """Parses the given ``key=value`` pairs and adds them to the given dict.

    .. note::

        This happens directly on the ``add_to`` object, i.e. making use of the
        mutability of the given dict. This function has no return value!

    Args:
        *pairs: Sequence of key=value strings
        add_to (dict): The dict to add the pairs to
        _log (TYPE, optional): A logger-like object
        attempt_conversion (bool, optional): Whether to attempt converting the
            strings to bool, float, int, and other types.
        **conversion_kwargs: Passed on to the conversion function,
            :py:func:`~utopya_cli._utils.convert_value`

    """

    _log.remark(
        "Parsing %d key-value pair%s ...",
        len(pairs),
        "s" if len(pairs) != 1 else "",
    )

    # Go over all pairs and add them to the given base dict
    for kv in pairs:
        key, val = kv.split("=", 1)

        # Process the key and traverse through the dict, already creating new
        # entries if needed. The resulting `d` will be the dict where the value
        # is written to (or deleted from).
        key_sequence = key.split(".")
        traverse_keys, last_key = key_sequence[:-1], key_sequence[-1]

        d = add_to
        for _key in traverse_keys:
            if _key not in d:
                d[_key] = dict()

            d = d[_key]

        # Process the value
        if attempt_conversion:
            val = convert_value(val, **conversion_kwargs)

        _log.remark("  %s  \t->   %s: %s", kv, ".".join(key_sequence), val)

        # Write or delete the entry
        if val is not DEL:
            if not isinstance(val, dict):
                d[last_key] = val

            else:
                from utopya.tools import recursive_update

                d[last_key] = recursive_update(
                    copy.deepcopy(d.get(last_key, {})),
                    val,
                )

        else:
            if last_key not in d:
                continue
            del d[last_key]


# -----------------------------------------------------------------------------
# Custom argument parsing


def parse_run_and_plots_cfg(
    *,
    model: "utopya.model.Model",
    run_cfg: str,
    plots_cfg: str,
    cfg_set: str,
    _interactive_mode: bool = False,
    _log=log,
) -> Tuple[str, str]:
    """Extracts paths to the run configuration and plots configuration by
    looking at the given arguments and the model's configuration sets.

    If ``_interactive_mode`` is given, will not read the run configuration but
    only the plots configuration where it would be confusing to have the
    corresponding log message appear. Also, in interactive mode, this will not
    lead to system exit if parsing failed.
    """
    if cfg_set and (run_cfg is None or plots_cfg is None):
        _log.info("Looking up config set '%s' ...", cfg_set)
        try:
            cfg_set = model.get_config_set(cfg_set)

        except ValueError as err:
            if _interactive_mode:
                raise
            _log.error(err)
            sys.exit(1)

        # Explicitly given arguments take precedence. Also, the config set may
        # not contain a run or eval configuration.
        if run_cfg is None and cfg_set.get("run"):
            if not _interactive_mode:
                run_cfg = cfg_set["run"]
                _log.note("  Using run.yml from config set.")
            else:
                _log.remark(
                    "  Not using run.yml in interactive plotting mode."
                )

        if plots_cfg is None and cfg_set.get("eval"):
            plots_cfg = cfg_set["eval"]
            _log.note("  Using eval.yml from config set.")

    return run_cfg, plots_cfg


def parse_update_dicts(
    *, _mode: str, _log=log, **all_arguments
) -> Tuple[dict, dict]:
    """Parses the given arguments, extracting update dictionaries for the
    Multiverse and the plots configuration

    Args:
        _mode (str): The mode argument. Can be ``run`` or ``eval``

    Returns:
        Tuple[dict, dict]: Multiverse update config and plots update config
    """
    from types import SimpleNamespace

    from utopya.tools import add_item

    # Make attribute access possible for arguments by using SimpleNamespace
    args = SimpleNamespace(**all_arguments)

    # To-be-populated update dicts:
    update_dict = {}
    update_plots_cfg = {}

    # . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
    # TODO Take into account framework-level defaults
    # TODO Make use of additional debug levels to select more verbose output
    add_item(
        args.debug,
        add_to=update_dict,
        key_path=("debug_level",),
    )
    if args.debug >= 1:
        # Set model log level to DEBUG and let PlotManager and WorkerManager
        # raise exceptions
        add_item(
            "debug",
            add_to=update_dict,
            key_path=("parameter_space", "log_levels", "model"),
        )
        add_item(
            True,
            add_to=update_dict,
            key_path=("plot_manager", "raise_exc"),
        )
        add_item(
            "raise",
            add_to=update_dict,
            key_path=("worker_manager", "nonzero_exit_handling"),
        )

    # . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
    # Arguments relevant only in run mode
    if _mode == "run":
        if args.note:
            add_item(
                args.note, add_to=update_dict, key_path=("paths", "model_note")
            )

        # TODO Replacement?
        # if args.sim_errors:
        #     add_item(
        #         args.sim_errors,
        #         add_to=update_dict,
        #         key_path=["worker_manager", "nonzero_exit_handling"],
        #     )

        if args.validate is not None:
            add_item(
                args.validate,
                add_to=update_dict,
                key_path=("perform_validation",),
            )

        if args.num_steps is not None:
            add_item(
                args.num_steps,
                add_to=update_dict,
                key_path=("parameter_space", "num_steps"),
            )

        if args.write_every is not None:
            add_item(
                args.write_every,
                add_to=update_dict,
                key_path=("parameter_space", "write_every"),
            )

        if args.write_start is not None:
            add_item(
                args.write_start,
                add_to=update_dict,
                key_path=("parameter_space", "write_start"),
            )

        if args.num_seeds is not None:
            add_item(
                args.num_seeds,
                value_func=lambda v: psp.ParamDim(default=42, range=[v]),
                add_to=update_dict,
                key_path=("parameter_space", "seed"),
                is_valid=lambda v: bool(v >= 1),
                ErrorMsg=lambda v: ValueError(
                    f"Argument --num-seeds needs to be ≥ 1, was {v}."
                ),
            )
            add_item(True, add_to=update_dict, key_path=("perform_sweep",))

        if args.num_workers is not None:
            add_item(
                args.num_workers,
                add_to=update_dict,
                key_path=("worker_manager", "num_workers"),
            )

        if args.run_mode is not None:
            add_item(
                args.run_mode == "sweep",
                add_to=update_dict,
                key_path=("perform_sweep",),
            )

        if args.worker_perform_task is not None:
            add_item(
                args.worker_perform_task,
                add_to=update_dict,
                key_path=("worker_kwargs", "perform_task"),
            )
            if (
                not args.worker_perform_task
                and args.perform_eval is not None
                and args.perform_eval
            ):
                raise ValueError("Cannot perform eval after --no-work.")
            if not args.worker_perform_task and args.perform_eval is None:
                raise NotImplementedError(
                    "Please explicityly specify "
                    "'--no-eval' option together with '--no-work'. Sorry for "
                    "the inconvenience."
                )

        if args.set_model_params:
            # TODO More elegant solution?
            if not update_dict.get("parameter_space"):
                update_dict["parameter_space"] = dict()

            if not update_dict["parameter_space"].get(args.model_name):
                update_dict["parameter_space"][args.model_name] = dict()

            set_entries_from_kv_pairs(
                *args.set_model_params,
                add_to=update_dict["parameter_space"][args.model_name],
                _log=_log,
            )

        if args.set_pspace_params:
            if not update_dict.get("parameter_space"):
                update_dict["parameter_space"] = dict()

            set_entries_from_kv_pairs(
                *args.set_pspace_params,
                add_to=update_dict["parameter_space"],
                _log=_log,
            )

    # . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
    # Arguments relevant *only* in eval mode
    elif _mode == "eval":
        # Can add eval-specific arugments here
        pass

    else:
        raise ValueError(f"Bad mode '{mode}'! Needs be: run or eval")

    # . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
    # Evaluate arguments that apply to both run and eval modes
    if args.load_parallel is not None:
        add_item(
            args.load_parallel,
            add_to=update_dict,
            key_path=("data_manager", "load_cfg", "data", "parallel"),
        )

    if args.cluster_mode is not None:
        add_item(
            args.cluster_mode,
            add_to=update_dict,
            key_path=("cluster_mode",),
        )

    if args.set_params:
        set_entries_from_kv_pairs(
            *args.set_params,
            add_to=update_dict,
            _log=_log,
        )

    if args.update_plots_cfg:
        try:
            set_entries_from_kv_pairs(
                *args.update_plots_cfg,
                add_to=update_plots_cfg,
                _log=_log,
            )

        except ValueError:
            from utopya.yaml import load_yml

            update_plots_cfg = load_yml(*args.update_plots_cfg)
            # FIXME needs to cover case where update_plots_cfg has more than
            #       a single argument

    return update_dict, update_plots_cfg



"""Signal handling when using the utopya backend, e.g. for catching stop
conditions and handling them gracefully."""

import signal
import time

SIG_STOPCOND = signal.SIGUSR1
"""Which signal to look out for if a stop condition was fulfilled. This should
match :py:data:`utopya.stop_conditions.SIG_STOPCOND`."""

SIGNAL_INFO = dict(got_signal=False, signum=None, frame=None, at_time=None)
"""A dict that holds information on whether any kind of signal was received
and at which time. This can be analysed by other modules to determine which
action to take."""


def _handle_signal(signum, frame):
    """A signal handler function that writes information into the
    ``SIGNAL_INFO`` dict."""
    SIGNAL_INFO["got_signal"] = True
    SIGNAL_INFO["signum"] = signum
    SIGNAL_INFO["frame"] = frame
    SIGNAL_INFO["at_time"] = time.time()


def attach_signal_handlers(
    *, for_stop_conds: bool = True, for_interrupts: bool = True
):
    """A function that can be invoked to attach signal handlers for use within
    utopya. There are two kinds of signals:

    - stop conditions, as implemented in :py:mod:`utopya.stop_conditions`
    - interrupt signals ``SIGTERM`` and ``SIGINT``

    Args:
        for_stop_conds (bool, optional): Whether to attach signal handlers for
            stop conditions.
        for_interrupts (bool, optional): Whether to attach signal handlers for
            interrupts.
    """
    if for_stop_conds:
        signal.signal(SIG_STOPCOND, _handle_signal)

    if for_interrupts:
        signal.signal(signal.SIGINT, _handle_signal)
        signal.signal(signal.SIGTERM, _handle_signal)



"""Tests `utopya batch` subcommand"""

import os
import shutil

import pytest

import utopya

from .. import get_cfg_fpath
from . import invoke_cli

BATCH_FILE: str = get_cfg_fpath("batch_file.yml")
"""Path to a test batch file"""

# -- Fixtures -----------------------------------------------------------------
from .._fixtures import *
from ..test_batch import skip_if_on_macOS


@pytest.fixture(autouse=True)
def register_test_project(tmp_projects):
    """Use on all tests in this module"""
    pass


# -----------------------------------------------------------------------------


@skip_if_on_macOS
def test_batch(with_test_models):  # FIXME creates test artifacts in output dir
    """Tests the `utopya batch` subcommand"""
    # Make sure the ExtendedModel has already run
    model = utopya.Model(name="ExtendedModel")
    mv = model.create_mv()
    mv.run()

    # Now actually perform the batch eval
    res = invoke_cli(("batch", "-d", "-s", "--note", "some_note", BATCH_FILE))
    print(res.output)

    assert res.exit_code == 0
    assert "time_series/mean_state" in res.output



"""Implements logging-related infrastructure"""

import logging as _logging
from typing import Dict

# -----------------------------------------------------------------------------

LOG_LEVELS: Dict[str, int] = {
    "trace": 5,
    "debug": _logging.DEBUG,
    "info": _logging.INFO,
    "warn": _logging.WARN,
    "warning": _logging.WARNING,
    "error": _logging.ERROR,
    "critical": _logging.CRITICAL,
    "fatal": _logging.FATAL,
    "not_set": _logging.NOTSET,
    "notset": _logging.NOTSET,
    "none": _logging.NOTSET,
}
"""A map of log level names to actual level values"""

# TODO add trace level and other intermediate levels?

DEFAULT_LOG_FORMAT = "%(levelname)-7s %(message)s"
"""The default logging format to use; can also include ``%(name)-14s`` here to
show the logger's name."""

_logging.basicConfig(
    format=DEFAULT_LOG_FORMAT,
    level=LOG_LEVELS["info"],
)

backend_logger = _logging.getLogger()
"""A backend-wide logger instance which is the same as the root logger.

.. note::

    The :py:class:`~utopya_backend.model.base.BaseModel` may adjust the level
    of this logger.
"""

# -----------------------------------------------------------------------------


def get_level(s: str) -> int:
    """Returns the integer log level from a string, looking it up in
    :py:data:`~utopya_backend.logging.LOG_LEVELS`.

    Args:
        s (str): Name of the log level, not case-sensitive.

    Returns:
        int: The desired log level
    """
    return LOG_LEVELS[s.lower()]



"""Tests the utopya run CLI command"""

import logging
import os
import time
import traceback

import pytest

from .. import ADVANCED_MODEL, DUMMY_MODEL
from .._fixtures import *
from . import invoke_cli


def _check_result(res, expected_exit: int = 0):
    if res.exception:
        print(res.output)
        traceback.print_tb(res.exception.__traceback__)

    elif res.exit_code != expected_exit:
        print(res.output)

    assert res.exit_code == expected_exit


# -----------------------------------------------------------------------------


def test_run(with_test_models, tmp_output_dir):
    """Tests the invocation of the utopya run command"""
    # Simplest case with the dummy model
    res = invoke_cli(("run", DUMMY_MODEL, "-d"))
    _check_result(res, expected_exit=0)
    assert not "Now creating plots" in res.output  # evaluation not attempted

    # Again, but with the advanced model
    res = invoke_cli(("run", ADVANCED_MODEL, "-d"))
    _check_result(res, expected_exit=0)

    # ... and with higher debug level
    res = invoke_cli(("run", ADVANCED_MODEL, "-dd"))
    _check_result(res, expected_exit=0)

    # Adjusting some of the meta config parameters, testing if they show up
    args = ("run", DUMMY_MODEL, "--no-eval", "--note", "some_note", "-d")
    res = invoke_cli(
        args
        + (
            "-N",
            "23",
            "--we",
            "5",
            "--ws",
            "7",
            "--mp",
            "foo.bar=ABCXYZ",
            "-W",
            "1",
        )
    )
    _check_result(res, expected_exit=0)

    assert "Updates to meta configuration" in res.output
    assert "ABCXYZ" in res.output


def test_run_existing(with_test_models, tmp_output_dir):
    """Tests the invocation of the utopya run_existing command"""

    # Run the dummy model with --no-work flag
    res_prep = invoke_cli(
        (
            "run",
            DUMMY_MODEL,
            "-d",
            "--num-seeds",
            "4",
            "--no-work",
            "--no-eval",
        )
    )
    _check_result(res_prep, expected_exit=0)
    assert "Finished working." in res_prep.output

    # search output for run directory
    __find = res_prep.output.find(tmp_output_dir)
    __end = __find + res_prep.output[__find:].find("\n")
    run_dir = res_prep.output[__find:__end]

    # check that all folders exist
    assert os.path.isdir(run_dir)
    assert os.path.isdir(os.path.join(run_dir, "config"))
    assert os.path.isdir(os.path.join(run_dir, "data"))
    assert os.path.isdir(os.path.join(run_dir, "eval"))

    # check that every universe output directory has a config
    # but no data and log
    for uni in range(1, 5):
        assert os.path.isdir(os.path.join(run_dir, "data", f"uni{uni}"))
        assert not os.path.isfile(
            os.path.join(run_dir, "data", f"uni{uni}", "data.h5")
        )
        assert not os.path.isfile(
            os.path.join(run_dir, "data", f"uni{uni}", "out.log")
        )

    # Repeat uni1 with run-existing
    res = invoke_cli(("run-existing", DUMMY_MODEL, run_dir, "--uni", "uni1"))
    _check_result(res, expected_exit=0)
    assert "uni1" in res.output
    assert "Now creating plots" not in res.output  # evaluation not attempted

    # Check that cannot be repeated again as data already exists
    res_fail_repeat = invoke_cli(
        ("run-existing", DUMMY_MODEL, run_dir, "--uni", "uni1")
    )
    _check_result(res_fail_repeat, expected_exit=1)

    # Repeat with uni2 and uni3
    res = invoke_cli(
        (
            "run-existing",
            DUMMY_MODEL,
            run_dir,
            "--uni",
            "uni2",
            "-u",
            "uni3",
        )
    )
    _check_result(res, expected_exit=0)
    assert run_dir in res.output
    assert "Now creating plots" not in res.output  # evaluation not attempted

    # Check that data exists
    for uni in range(1, 4):
        assert os.path.isfile(
            os.path.join(run_dir, "data", f"uni{uni}", "data.h5")
        )
        assert os.path.isfile(
            os.path.join(run_dir, "data", f"uni{uni}", "out.log")
        )

    # Check that uni4 never was run
    assert not os.path.isfile(os.path.join(run_dir, "data", "uni4", "data.h5"))
    assert not os.path.isfile(os.path.join(run_dir, "data", "uni4", "out.log"))

    # Repeat uni1 with clear existing option
    bad_file_path = os.path.join(
        run_dir, "data", "uni1", "this_file_should_not_exist.txt"
    )
    bad_file = open(bad_file_path, "a")
    bad_file.write("New line of text")
    bad_file.close()

    res = invoke_cli(
        (
            "run-existing",
            DUMMY_MODEL,
            run_dir,
            "--uni",
            "uni1",
            "--clear-existing",
        )
    )
    _check_result(res, expected_exit=0)
    assert "uni1" in res.output
    assert "Now creating plots" not in res.output  # evaluation not attempted

    assert not os.path.isfile(
        os.path.join(run_dir, "data", "uni1", "this_file_should_not_exist.txt")
    )

    # Run all but skip existing
    res = invoke_cli(
        (
            "run-existing",
            DUMMY_MODEL,
            run_dir,
            "--skip-existing",
        )
    )
    _check_result(res, expected_exit=0)

    # Check that uni4 was run
    assert "Finished working. Total tasks worked on: 1" in res.output
    assert os.path.isfile(os.path.join(run_dir, "data", "uni4", "data.h5"))
    assert os.path.isfile(os.path.join(run_dir, "data", "uni4", "out.log"))

    # Re-run all with clear existing option
    res = invoke_cli(
        (
            "run-existing",
            DUMMY_MODEL,
            run_dir,
            "--clear-existing",
        )
    )
    _check_result(res, expected_exit=0)

    assert "Finished working. Total tasks worked on: 4" in res.output


def test_eval(with_test_models, tmp_output_dir, delay):
    """Tests the invocation of the utopya eval command"""
    # Simplest case
    res = invoke_cli(("eval", ADVANCED_MODEL, "-d"))
    _check_result(res, expected_exit=0)

    time.sleep(1)

    # Adjusting some of the meta config parameters
    args = ("eval", ADVANCED_MODEL, "-d")
    res = invoke_cli(args + ("-p", "plot_manager.raise_exc=true"))
    _check_result(res, expected_exit=0)
    assert "Updates to meta configuration" in res.output



"""The :py:mod:`utopya_backend` package is a standalone package that provides
tools for implementation of Python-based models that use :py:mod:`utopya` as a
simulation management frontend.

Refer to the :ref:`the user manual <backend_overview>` for more information.
"""

# isort: skip_file

from .logging import *
from .model import BaseModel, StepwiseModel
from .tools import import_package_from_dir, load_cfg_file
from .benchmark import ModelBenchmarkMixin



"""Tests the benchmark mixin"""

import time

import pytest

from utopya.tools import write_yml
from utopya_backend import ModelBenchmarkMixin
from utopya_backend.benchmark import (
    DEFAULT_TIMERS_CUMULATIVE,
    DEFAULT_TIMERS_ONE_SHOT,
    Timer,
)

from .._fixtures import *
from .test_model import MyModel, MyStepwiseModel, minimal_pspace_cfg

# -----------------------------------------------------------------------------


class BenchmarkedBaseModel(ModelBenchmarkMixin, MyModel):
    def should_write(self):
        return True


class BenchmarkedModel(ModelBenchmarkMixin, MyStepwiseModel):
    pass


# -----------------------------------------------------------------------------


def test_Timer_init():
    t = Timer("foo")
    assert t.name == "foo"
    assert not t.running
    assert not t.finished
    assert t.elapsed == 0.0
    assert t.one_shot is False

    assert "Timer 'foo'" in str(t)
    assert "0s elapsed" in str(t)
    assert "running" not in str(t)
    assert "finished" not in str(t)

    # One-shot timer
    t = Timer("bar", one_shot=True)
    assert t.name == "bar"
    assert not t.running
    assert not t.finished
    assert t.elapsed == 0.0
    assert t.one_shot is True

    # Custom time function
    t = Timer("bar", time_func=time.time)


def test_timer_operation():
    DT = 0.05
    match_is_finished = "was already marked as finished"
    match_is_one_shot = "is a one-shot timer"
    match_is_paused = "Cannot pause already paused timer"

    def sleep(dt=DT):
        time.sleep(dt)

    # Regular operation
    t = Timer("foo")
    assert t.elapsed == 0.0
    assert not t.running

    t.start()
    assert t.running
    sleep()

    assert "running" in str(t)
    assert "finished" not in str(t)

    # Can get elapsed time without pausing
    t1 = t.elapsed
    assert t1 >= DT

    # After pausing, timer does not continue
    t.pause()
    assert not t.running

    t2 = t.elapsed
    assert t2 >= t1
    sleep()
    assert t.elapsed == t2

    # Can unpause to continue
    t.unpause()
    sleep()
    t3 = t.elapsed
    assert t3 > t2 + DT
    assert t.running

    # pause has a return value
    t4 = t.pause()
    assert not t.running
    sleep()

    # can stop a paused timer (and also have a return value)
    assert t.stop() == t4
    assert t.finished

    # Can directly start a timer
    t = Timer("bar", start=True)
    assert t.elapsed > 0.0
    t.pause()

    # Cannot pause a paused timer
    with pytest.raises(RuntimeError, match=match_is_paused):
        t.pause()

    # Cannot modify a finished timer
    t.stop()
    assert t.elapsed > 0.0
    assert t.finished
    assert not t.running

    assert "running" not in str(t)
    assert "finished" in str(t)

    with pytest.raises(RuntimeError, match=match_is_finished):
        t.start()

    with pytest.raises(RuntimeError, match=match_is_finished):
        t.unpause()

    with pytest.raises(RuntimeError, match=match_is_finished):
        t.pause()

    with pytest.raises(RuntimeError, match=match_is_finished):
        t.stop()

    # Cannot unpause or restart a one-shot timer
    t = Timer("baz", one_shot=True, start=True)
    assert t.running
    assert t.elapsed > 0.0
    assert not t.finished
    t.pause()

    with pytest.raises(RuntimeError, match=match_is_one_shot):
        t.unpause()

    with pytest.raises(RuntimeError, match=match_is_finished):
        t.start()


def test_ModelBenchmarkMixin(minimal_pspace_cfg, tmpdir):
    cfg = minimal_pspace_cfg
    cfg["root_model_name"] = "MyStepwiseModel"
    cfg["MyStepwiseModel"] = dict(
        sleep_time=0.01,
        benchmark=dict(enabled=True, add_to_monitor=True),
    )
    cfg["monitor_emit_interval"] = 0.1
    cfg["num_steps"] = 42

    cfg_path = tmpdir.join("cfg.yml")
    write_yml(cfg, path=cfg_path)

    # Initialize it, running init and setup
    m = BenchmarkedModel(cfg_file_path=cfg_path)

    print("one shot:", m.elapsed_one_shot)
    print("cumulative:", m.elapsed_cumulative)
    print("\n".join(f"{n}: {t}" for n, t in m.timers.items()))

    INIT_TIMERS = ("init", "setup", "monitor")
    assert all(not t.running for n, t in m.timers.items() if n != "simulation")
    assert all(
        not t.finished for n, t in m.timers.items() if n not in INIT_TIMERS
    )
    assert all(
        t.elapsed == 0
        for n, t in m.timers.items()
        if n not in INIT_TIMERS + ("simulation",)
    )

    assert m.timers["simulation"].running
    assert m.timers["simulation"].elapsed > 0

    # Key errors
    with pytest.raises(ValueError, match="No benchmark timer named"):
        m.start_timer("i do not exist")

    # Let it run and check timers again
    m.run()

    print("one shot:", m.elapsed_one_shot)
    print("cumulative:", m.elapsed_cumulative)
    print("\n".join(f"{n}: {t}" for n, t in m.timers.items()))

    # ... all should have run and be finished now (except teardown)
    timers = [t for n, t in m.timers.items() if n != "teardown"]
    assert all(not t.running for t in timers)
    assert all(t.finished for t in timers)
    assert all(t.elapsed > 0 for t in timers)

    # Delete the model
    timers = m.timers
    del m
    assert timers["teardown"].elapsed > 0.0
    assert timers["teardown"].finished


def test_ModelBenchmarkMixin_in_basemodel(minimal_pspace_cfg, tmpdir):
    cfg = minimal_pspace_cfg
    cfg["root_model_name"] = "MyModel"
    cfg["MyModel"] = dict(
        some_param="foo", max_n_iter=13, benchmark=dict(enabled=True)
    )

    cfg_path = tmpdir.join("cfg.yml")
    write_yml(cfg, path=cfg_path)

    # Initialize it, running init and setup
    m = BenchmarkedBaseModel(cfg_file_path=cfg_path)

    print("one shot:", m.elapsed_one_shot)
    print("cumulative:", m.elapsed_cumulative)
    print("\n".join(f"{n}: {t}" for n, t in m.timers.items()))

    INIT_TIMERS = ("init", "setup", "monitor")
    assert all(not t.running for n, t in m.timers.items() if n != "simulation")
    assert all(
        not t.finished for n, t in m.timers.items() if n not in INIT_TIMERS
    )
    assert all(
        t.elapsed == 0
        for n, t in m.timers.items()
        if n not in INIT_TIMERS + ("simulation",)
    )

    assert m.timers["simulation"].running
    assert m.timers["simulation"].elapsed > 0

    # Key errors
    with pytest.raises(ValueError, match="No benchmark timer named"):
        m.start_timer("i do not exist")

    # Let it run and check timers again
    m.run()

    print("one shot:", m.elapsed_one_shot)
    print("cumulative:", m.elapsed_cumulative)
    print("\n".join(f"{n}: {t}" for n, t in m.timers.items()))

    # ... all should have run and be finished now (except teardown)
    timers = [t for n, t in m.timers.items() if n not in ("teardown",)]
    assert all(not t.running for t in timers)
    assert all(t.finished for t in timers)
    assert all(t.elapsed > 0 for t in timers)

    # Delete the model
    timers = m.timers
    del m
    assert timers["teardown"].elapsed > 0.0
    assert timers["teardown"].finished


def test_ModelBenchmarkMixin_disabled(minimal_pspace_cfg, tmpdir):
    cfg = minimal_pspace_cfg
    cfg["root_model_name"] = "MyStepwiseModel"
    cfg["MyStepwiseModel"] = dict(
        sleep_time=0.01, benchmark=dict(enabled=False)
    )
    cfg["monitor_emit_interval"] = 0.1
    cfg["num_steps"] = 42

    cfg_path = tmpdir.join("cfg.yml")
    write_yml(cfg, path=cfg_path)

    # Initialize it
    m = BenchmarkedModel(cfg_file_path=cfg_path)
    print(m.elapsed)
    print("\n".join(f"{n}: {t}" for n, t in m.timers.items()))

    # All timers should be reset.
    assert all(not t.running for t in m.timers.values())
    assert all(not t.finished for t in m.timers.values())
    assert all(t.elapsed == 0 for t in m.timers.values())

    # Let it run and check timers again
    m.run()
    print(m.elapsed)
    print("\n".join(f"{n}: {t}" for n, t in m.timers.items()))

    # None should have run
    assert all(not t.running for t in m.timers.values())
    assert all(not t.finished for t in m.timers.values())
    assert all(t.elapsed == 0 for t in m.timers.values())

    # Delete the model
    timers = m.timers
    del m
    assert timers["teardown"].elapsed == 0.0
    assert not timers["teardown"].finished



"""This module provides graph plotting functions.

.. default-domain:: utopya.eval.plots
"""

import copy
import logging
import warnings
from itertools import chain, product
from typing import Callable, Sequence, Union

import networkx as nx
import numpy as np
import xarray as xr
from dantro.tools import adjusted_log_levels as _adjusted_log_levels

from . import PlotHelper, is_plot_func
from ._graph import GraphPlot
from ._mpl import ColorManager

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------
# Helper Methods


def graph_array_from_group(
    graph_group: "utopya.eval.groups.GraphGroup",
    *,
    graph_creation: dict = None,
    register_property_maps: dict = None,
    clear_existing_property_maps: bool = True,
    times: dict = None,
    sel: dict = None,
    isel: dict = None,
) -> xr.DataArray:
    """From a :py:class:`utopya.eval.groups.GraphGroup` creates a
    :py:class:`xarray.DataArray` with ``object`` dtype, containing the
    :py:class:`networkx.Graph` objects created from the graph group at the
    specified points in the group's coordinate space.

    From all coordinates provided via the selection kwargs the cartesian
    product is taken. Each of those points represents one entry in the returned
    :py:class:`xarray.DataArray`.
    The selection kwargs in ``graph_creation`` are ignored silently.

    Args:
        graph_group (utopya.eval.groups.GraphGroup): The graph group to create
            the graph array from
        graph_creation (dict, optional): Graph creation configuration
        register_property_maps (dict, optional): Property maps to be registered
            before graph creation
        clear_existing_property_maps (bool, optional): Whether to remove any
            existing property map at first
        times (dict, optional): Equivalent to a ``sel.time`` entry.
            (**Deprecated!**)
        sel (dict, optional): Select by value
        isel (dict, optional): Select by index

    Returns:
        xarray.DataArray: :py:class:`networkx.Graph` with the respective graph
            group coordinates
    """
    # Remove all selection kwargs from the configuration
    graph_creation = (
        copy.deepcopy(graph_creation) if graph_creation is not None else {}
    )
    graph_creation.pop("at_time", None)
    graph_creation.pop("at_time_idx", None)
    graph_creation.pop("sel", None)
    graph_creation.pop("isel", None)

    # Parse the selectors
    sel = sel if sel else {}
    isel = isel if isel else {}

    # Any ``times`` entry is transformed into the respective sel/isel entry.
    # TODO The ``times`` argument is deprecated. Remove it eventually.
    if times is not None:
        warnings.warn(
            "The 'times' argument is deprecated and will be removed. Use the "
            "'sel' and 'isel' arguments instead to specify a time selection.",
            DeprecationWarning,
        )

        if len(times) > 1:
            raise TypeError(
                "Received ambiguous animation time specifications. Need "
                "_one_ of: from_property, sel , isel"
            )

        if "from_property" in times:
            sel["time"] = list(
                graph_group._get_item_or_pmap(times["from_property"])
                .coords["time"]
                .values
            )

        elif "sel" in times:
            sel["time"] = times["sel"]

        elif "isel" in times:
            isel["time"] = times["isel"]

    # If needed, extract coordinates from property data
    for dim in sel:
        if isinstance(sel[dim], dict) and "from_property" in sel[dim]:
            _k = sel[dim]["from_property"]
            log.note(
                "Extracting graph animation selection for dimension '%s' "
                "from property '%s' ...",
                dim,
                _k,
            )

            _vals = graph_group._get_item_or_pmap(_k).coords[dim]

            # subselection
            _isel = sel[dim].get("isel")  # for subselection
            _sel = sel[dim].get("sel")  # for subselection

            if _isel:
                log.remark("Performing index-based subselection:  %s", _isel)
                _vals = _vals.isel(_isel)
            if _sel:
                log.remark("Performing value-based subselection:  %s", _sel)
                _vals = _vals.sel(_sel)

            # assign selector values
            log.remark("Selector for '%s' dimension:\n%s", dim, _vals)
            sel[dim] = list(_vals.values)

    # With the given selectors, can now set up the graph-DataArray and create
    # a graph from each coordinate combination.
    # In order to take the cartesion product of all (sel and isel) selectors,
    # collect them all and separate them again afterwards.
    dims = list(sel.keys()) + list(isel.keys())
    coords = {d: [] for d in dims}

    all_selectors = [
        c if isinstance(c, (list, tuple, np.ndarray)) else [c]
        for c in chain(sel.values(), isel.values())
    ]

    graphs = np.empty(tuple(len(c) for c in all_selectors), dtype="object")

    indices = [[i for i in range(len(c))] for c in all_selectors]

    N = len(list(product(*indices)))
    log.note("Creating %d graphs for animation ...", N)

    for i, (idxs, selector) in enumerate(
        zip(product(*indices), product(*all_selectors))
    ):
        sel_coords = selector[: len(sel)]
        isel_coords = selector[len(sel) :]

        # Prepare the selectors for the single points in coordinate space
        select = {d: c for d, c in zip(sel.keys(), sel_coords)}
        iselect = {d: c for d, c in zip(isel.keys(), isel_coords)}

        with _adjusted_log_levels(("dantro.groups.graph", 21)):
            g = GraphPlot.create_graph_from_group(
                graph_group=graph_group,
                register_property_maps=register_property_maps,
                clear_existing_property_maps=clear_existing_property_maps,
                sel=select,
                isel=iselect,
                **graph_creation,
            )

        graphs[idxs] = g

        # Extract the coordinate values from the networkx graph attributes
        # NOTE This assumes that there are no additional graph attributes that
        #      also match a dimension name ...
        for d, c in g.graph.items():
            if d not in coords:
                continue
            if coords[d] == [] or coords[d][-1] != c:
                coords[d].append(c)

        print(f"  {i+1:4d} / {N}:  {g}", end="\r")

    coords = {d: coords[d] for d in dims}

    # Combine the data and coordinate information in a DataArray
    graphs = xr.DataArray(graphs, dims=dims, coords=coords)

    return graphs


def graph_animation_update(
    *,
    hlpr: PlotHelper,
    graphs: xr.DataArray = None,
    graph_group=None,
    graph_creation: dict = None,
    register_property_maps: dict = None,
    clear_existing_property_maps: bool = True,
    positions: dict = None,
    animation_kwargs: dict = None,
    suptitle_kwargs: dict = None,
    suptitle_formatter: Callable = None,
    **drawing_kwargs,
):
    """Graph animation frame generator. Yields whenever the plot helper may
    grab the current frame.

    If ``graphs`` is given, the networkx graphs in the array are used to create
    the frames.

    Otherwise, use a graph group. The frames are defined via the selectors in
    ``animation_kwargs``. From all provided coordinates the cartesian product
    is taken. Each of those points defines one graph and thus one frame.
    The selection kwargs in ``graph_creation`` are ignored silently.

    Args:
        hlpr (PlotHelper): The plot helper
        graphs (xarray.DataArray, optional): Networkx graphs to draw. The array
            will be flattened beforehand.
        graph_group (None, optional): Required if ``graphs`` is None. The
            GraphGroup from which to generate the animation frames as specified
            via sel and isel in ``animation_kwargs``.
        graph_creation (dict, optional): Graph creation configuration. Passed
            to :py:meth:`~utopya.eval.plots._graph.GraphPlot.create_graph_from_group`
            if ``graph_group`` is given.
        register_property_maps (dict, optional): Passed to
            :py:meth:`~utopya.eval.plots._graph.GraphPlot.create_graph_from_group`
            if ``graph_group`` is given.
        clear_existing_property_maps (bool, optional): Passed to
            :py:meth:`~utopya.eval.plots._graph.GraphPlot.create_graph_from_group`
            if ``graph_group`` is given.
        positions (dict, optional): The node position configuration.
            If ``update_positions`` is True the positions are reconfigured for
            each frame.
        animation_kwargs (dict, optional): Animation configuration. The
            following arguments are allowed:

            times (dict, optional):
                *Deprecated*: Equivaluent to a ``sel.time`` entry.
            sel (dict, optional):
                Select by value. Coordinate values (or ``from_property`` entry)
                keyed by dimension name.
            isel (dict, optional):
                Select by index. Coordinate indices keyed by dimension. May be
                given together with ``sel`` if no key appears in both.
            update_positions (bool, optional):
                Whether to reconfigure the node positions for each frame
                (default=False).
            update_colormapping (bool, optional):
                Whether to reconfigure the nodes' and edges'
                :py:class:`~dantro.plot.utils.color_mngr.ColorManager` for
                each frame (default=False). If False, the colormapping (and the
                colorbar) is configured with the first frame and then fixed.
            skip_empty_frames (bool, optional):
                Whether to skip the frames where the selected graph is missing
                or of a type different than ``nx.Graph`` (default=False).
                If False, such frames are empty.

        suptitle_kwargs (dict, optional): Passed on to the
            :py:class:`~utopya.eval.plothelper.PlotHelper` and its
            ``set_suptitle`` helper function. Only used in animation mode.
            The ``title`` can be a format string containing a placeholder with
            the dimension name as key for each dimension along which selection
            is done. The format string is updated for each frame of the
            animation. The default is ``<dim-name> = {<dim-name>}`` for each
            dimension.
        **drawing_kwargs: Passed to
            :py:class:`~utopya.eval.plots._graph.GraphPlot`
    """
    suptitle_kwargs = (
        copy.deepcopy(suptitle_kwargs) if suptitle_kwargs is not None else {}
    )
    animation_kwargs = (
        copy.deepcopy(animation_kwargs) if animation_kwargs is not None else {}
    )
    update_positions = animation_kwargs.pop("update_positions", False)
    update_colormapping = animation_kwargs.pop("update_colormapping", False)
    skip_empty_frames = animation_kwargs.pop("skip_empty_frames", False)
    sel = animation_kwargs.pop("sel", None)
    isel = animation_kwargs.pop("isel", None)

    if graphs is None:
        graphs = graph_array_from_group(
            graph_group=graph_group,
            graph_creation=graph_creation,
            register_property_maps=register_property_maps,
            clear_existing_property_maps=clear_existing_property_maps,
            sel=sel,
            isel=isel,
            **animation_kwargs,
        )

    else:
        # Apply selectors to the graphs DataArray
        if sel is not None:
            graphs = graphs.sel(**sel)
        if isel is not None:
            graphs = graphs.isel(**isel)

    # Prepare the suptitle format string once for all frames
    if "title" not in suptitle_kwargs:
        suptitle_kwargs["title"] = "; ".join(
            [f"{dim}" + " = {" + dim + "}" for dim in graphs.coords]
        )

    def get_graph_and_coords(graphs: xr.DataArray):
        """Generator that yields the (graph, coordinates) pairs"""
        dims = graphs.dims
        if dims:
            # Stack all dimensions of the xr.DataArray
            graphs = graphs.stack(_flat=dims)

            # For each entry, yield the graph and the coords
            for el in graphs:
                g = el.item()
                coords = {
                    d: c for d, c in zip(dims, el.coords["_flat"].item())
                }
                # Also get the scalar coordinates which were not stacked
                coords.update(
                    {d: c.item() for d, c in el.coords.items() if d != "_flat"}
                )
                yield g, coords

        else:
            # zero-dimensional: extract single graph item
            g = graphs.item()
            coords = {d: c.item() for d, c in graphs.coords.items()}
            yield g, coords

    # Don't show the axis. This is also done in GraphPlot.draw but need to do
    # it here in case the first frame is empty.
    hlpr.ax.axis("off")

    # Indicator for things to be done only once after the first GraphPlot.draw
    # call.
    _first_graph = True

    # Always configure the colormanagers on the first GraphPlot.draw call.
    # Overwritten by the update_colormapping kwarg afterwards.
    _update_colormapping = True

    # Loop over all remaining graphs
    for g, coords in get_graph_and_coords(graphs):

        _missing_val = not isinstance(g, nx.Graph)

        # On missing entries, skip the frame if skip_empty_frames=True.
        # If skip_empty_frames=False, the suptitle helper is applied but no
        # graph is drawn.
        if _missing_val and skip_empty_frames:
            continue

        if not _missing_val:
            # Use a new GraphPlot for the next frame such that the `select`
            # kwargs are re-evaluated.
            with _adjusted_log_levels(
                ("dantro.plot.utils.color_mngr", 15),
                ("utopya.eval.plots._graph", 15),
            ):
                log.remark(
                    "Drawing %s ...  (%s)",
                    g,
                    ", ".join(f"{k}: {v}" for k, v in coords.items()),
                )
                gp = GraphPlot(
                    g=g,
                    fig=hlpr.fig,
                    ax=hlpr.ax,
                    positions=positions,
                    **drawing_kwargs,
                )
                gp.draw(
                    suppress_cbar=not _update_colormapping,
                    update_colormapping=_update_colormapping,
                )

        # Update the suptitle format string and invoke the helper
        st_kwargs = copy.deepcopy(suptitle_kwargs)
        if suptitle_formatter is None:
            suptitle_formatter = lambda fstr, **coords: fstr.format(**coords)

        st_kwargs["title"] = suptitle_formatter(st_kwargs["title"], **coords)
        hlpr.invoke_helper("set_suptitle", **st_kwargs)

        if _first_graph:
            # Fix the y-position of the suptitle after the first invocation
            # since repetitive invocations of the set-suptitle helper would
            # re-adjust the subplots size each time. Use the matplotlib default
            # if not given.
            if "y" not in suptitle_kwargs:
                suptitle_kwargs["y"] = 0.98

            _update_colormapping = update_colormapping

        # Let the writer grab the current frame
        yield

        if not _missing_val:
            # Clean up the frame
            with _adjusted_log_levels(("utopya.eval.plots._graph", 15)):
                gp.clear_plot(keep_colorbars=not _update_colormapping)

            if _first_graph:
                # If the positions should be fixed, overwrite the positions arg
                if not update_positions:
                    positions = dict(from_dict=gp.positions)

                # Done handling the first GraphPlot.draw call
                _first_graph = False


# -----------------------------------------------------------------------------
# Plot functions


@is_plot_func(use_dag=True, supports_animation=True)
def draw_graph(
    *,
    hlpr: PlotHelper,
    data: dict,
    graph_group_tag: str = "graph_group",
    graph: Union[nx.Graph, xr.DataArray] = None,
    graph_creation: dict = None,
    graph_drawing: dict = None,
    graph_animation: dict = None,
    register_property_maps: Sequence[str] = None,
    clear_existing_property_maps: bool = True,
    suptitle_kwargs: dict = None,
    suptitle_formatter: Callable = None,
):
    """Draws a graph either from a :py:class:`~utopya.eval.groups.GraphGroup`
    or directly from a :py:class:`networkx.Graph` using the
    :py:class:`~utopya.eval.plots._graph.GraphPlot` class.

    If the graph object is to be created from a graph group the latter needs to
    be selected via the TransformationDAG. Additional property maps can also be
    made available for plotting, see ``register_property_map`` argument.
    Animations can be created either from a graph group by using the select
    interface in ``graph_animation`` or by passing a DataArray of networkx
    graphs via the ``graph`` argument.

    For more information on how to use the transformation framework, refer to
    the `dantro documentation <https://dantro.readthedocs.io/en/stable/plotting/plot_data_selection.html>`_
    and its :py:class:`~dantro.dag.TransformationDAG` class.

    For more information on how to configure the graph layout refer to the
    :py:class:`~utopya.eval.plots._graph.GraphPlot` documentation.

    Args:
        hlpr (PlotHelper): The PlotHelper instance for this plot
        data (dict): Data from TransformationDAG selection
        graph_group_tag (str, optional): The TransformationDAG tag of the graph
            group
        graph (Union[networkx.Graph, xarray.DataArray], optional): If given,
            the ``data`` and ``graph_creation`` arguments are ignored and this
            graph is drawn directly.
            If a DataArray of graphs is given, the first graph is drawn for a
            single graph plot. In animation mode the (flattened) array
            represents the animation frames.
        graph_creation (dict, optional): Configuration of the graph creation.
            Passed to ``GraphGroup.create_graph``.
        graph_drawing (dict, optional): Configuration of the graph layout.
            Passed to :py:class:`~utopya.eval.plots._graph.GraphPlot`.
        graph_animation (dict, optional): Animation configuration. The
            following arguments are allowed:

            times (dict, optional):
                *Deprecated*: Equivaluent to a ``sel.time`` entry.
            sel (dict, optional):
                Select by value. Dictionary with dimension names as keys. The
                values may either be coordinate values or a dict with a
                ``from_property`` (str) entry which specifies a container
                within the :py:class:`~utopya.eval.groups.GraphGroup` or
                registered external data from which the coordinates are
                extracted.
                Additionally, there can be ``isel`` and ``sel`` keys beside the
                ``from_property`` key, performing a subselection on those
                coordinates.
            isel (dict, optional):
                Select by index. Coordinate indices keyed by dimension. May be
                given together with ``sel`` if no key appears in both.
            update_positions (bool, optional):
                Whether to update the node positions for each frame by
                recalculating the layout with the parameters specified in
                graph_drawing.positions. If this parameter is not given or
                false, the positions are calculated once initially and then
                fixed.
            update_colormapping (bool, optional):
                Whether to reconfigure the nodes' and edges'
                :py:class:`~dantro.plot.utils.color_mngr.ColorManager` for
                each frame (default=False). If False, the colormapping (and the
                colorbar) is configured with the first frame and then fixed.
            skip_empty_frames (bool, optional):
                Whether to skip the frames where the selected graph is missing
                or of a type different than :py:class:`networkx.Graph`
                (default=False). If False, such frames are empty.

        register_property_maps (Sequence[str], optional): Names of properties
            to be registered in the graph group before the graph creation.
            The property names must be valid TransformationDAG tags, i.e., be
            available in ``data``. Note that the tags may not conflict with any
            valid path reachable from inside the selected
            :py:class:`~utopya.eval.groups.GraphGroup`.
        clear_existing_property_maps (bool, optional): Whether to clear any
            existing property maps from the selected
            :py:class:`~utopya.eval.groups.GraphGroup`. This is enabled by
            default to reduce side effects from previous plots. Set this to
            False if you have property maps registered with the GraphGroup
            that you would like to keep.
        suptitle_kwargs (dict, optional): Passed on to the PlotHelper's
            ``set_suptitle`` helper function. Only used in animation mode.
            The ``title`` can be a format string containing a placeholder with
            the dimension name as key for each dimension along which selection
            is done. The format string is updated for each frame of the
            animation. The default is ``<dim-name> = {<dim-name>}`` for each
            dimension.
        suptitle_formatter (Callable, optional): If given, this should be a
            function that takes coordinates and their values as keyword
            arguments and returns the string that is to be used as the title.
            Only used in animation mode.

    Raises:
        ValueError: On invalid or non-computed
            :py:class:`~dantro.dag.TransformationDAG` tags in
            ``register_property_maps`` or invalid graph group tag.
    """
    # Work on a copy such that the original configuration is not modified
    graph_drawing = copy.deepcopy(graph_drawing if graph_drawing else {})

    # Get the sub-configurations for the drawing of the graph
    select = graph_drawing.get("select", {})
    pos_kwargs = graph_drawing.get("positions", {})
    node_kwargs = graph_drawing.get("nodes", {})
    edge_kwargs = graph_drawing.get("edges", {})
    node_label_kwargs = graph_drawing.get("node_labels", {})
    edge_label_kwargs = graph_drawing.get("edge_labels", {})
    mark_nodes_kwargs = graph_drawing.get("mark_nodes", {})
    mark_edges_kwargs = graph_drawing.get("mark_edges", {})

    def get_dag_data(tag):
        try:
            return data[tag]
        except KeyError as err:
            _available_tags = ", ".join(data.keys())
            raise ValueError(
                f"No tag '{tag}' found in the data selected by the DAG! Make "
                "sure the tag is named correctly and is selected to be "
                "computed; adjust the 'compute_only' argument if needed."
                "\nThe following tags are available in the DAG results: "
                f"{_available_tags}"
            ) from err

    # Prepare graph group and external property data
    graph_group = get_dag_data(graph_group_tag) if graph is None else None
    property_maps = None
    if register_property_maps:
        property_maps = {}
        for tag in register_property_maps:
            property_maps[tag] = get_dag_data(tag)

    # If not in animation mode, make a single graph plot
    if not hlpr.animation_enabled:
        # Set up a GraphPlot instance
        if graph_group is not None:
            # Create GraphPlot from graph group
            gp = GraphPlot.from_group(
                graph_group=graph_group,
                graph_creation=graph_creation,
                register_property_maps=property_maps,
                clear_existing_property_maps=clear_existing_property_maps,
                fig=hlpr.fig,
                ax=hlpr.ax,
                **graph_drawing,
            )

        else:
            if graph_creation is not None:
                warnings.warn(
                    "Received both a 'graph' argument and a 'graph_creation' "
                    "configuration. The latter will be ignored. To remove "
                    "this warning set graph_creation to None.",
                    UserWarning,
                )

            if isinstance(graph, xr.DataArray):
                # Use the first array element for a single graph plot
                g = graph.values.flat[0]

                if graph.size > 1:
                    log.caution(
                        "Received a DataArray of size %d as 'graph' argument, "
                        "performing a single plot using the first entry. "
                        "Animations must be enabled by setting "
                        "animation.enabled to True.",
                        graph.size,
                    )

            else:
                g = graph

            # Create a GraphPlot from a nx.Graph
            gp = GraphPlot(g=g, fig=hlpr.fig, ax=hlpr.ax, **graph_drawing)

        # Make the actual plot via the GraphPlot
        gp.draw()

    # In animation mode, register the animation frame generator
    else:
        # Prepare animation kwargs for the update routine
        graph_animation = graph_animation if graph_animation else {}

        def update():
            """The animation frames generator.
            See :py:meth:`~utopya.eval.plots.dag.graph.graph_animation_update`.
            """
            if graph_group is None and not isinstance(graph, xr.DataArray):
                raise TypeError(
                    "Failed to create animation due to invalid type of the "
                    "'graph' argument. Required: xr.DataArray (filled "
                    f"with graph objects). Received: {type(graph)}"
                )

            yield from graph_animation_update(
                hlpr=hlpr,
                graphs=graph if isinstance(graph, xr.DataArray) else None,
                graph_group=graph_group,
                graph_creation=graph_creation,
                register_property_maps=property_maps,
                clear_existing_property_maps=clear_existing_property_maps,
                suptitle_kwargs=suptitle_kwargs,
                suptitle_formatter=suptitle_formatter,
                animation_kwargs=graph_animation,
                **graph_drawing,
            )

        hlpr.register_animation_update(update, invoke_helpers_before_grab=True)



"""This module implements the :py:class:`~utopya.stop_conditions.StopCondition`
class, which is used by the :py:class:`~utopya.workermanager.WorkerManager` to
stop a worker process in certain situations.

In addition, it implements a set of basic stop condition functions and
provides the :py:func:`~utopya.stop_conditions.stop_condition_function`
decorator which is required to make them accessible by name.
"""

import copy
import logging
import time
import warnings
from typing import Callable, Dict, List, Set, Tuple, Union

from dantro.data_ops.db import BOOLEAN_OPERATORS as _OPERATORS
from paramspace.tools import recursive_getitem as _recursive_getitem

log = logging.getLogger(__name__)

SIG_STOPCOND = "SIGUSR1"
"""Signal to use for stopping workers with fulfilled stop conditions"""

STOP_CONDITION_FUNCS: Dict[str, Callable] = dict()
"""Registered stop condition functions are stored in this dictionary. These
functions evaluate whether a certain stop condition is actually fulfilled.

To that end, a :py:class:`~utopya.task.WorkerTask` object is passed to these
functions, the information in which can be used to determine whether the
condition is fulfilled.
The signature of these functions is: ``(task: WorkerTask, **kws) -> bool``
"""

_FAILED_MONITOR_ENTRY_CHECKS = []
"""Keeps track of failed monitor entry checks in the
:py:func:`~utopya.stop_conditions.check_monitor_entry`
stop condition function in order to avoid repetitive warnings.
"""

# -----------------------------------------------------------------------------


class StopCondition:
    """A StopCondition object holds information on the conditions in which a
    worker process should be stopped.
    """

    def __init__(
        self,
        *,
        to_check: List[dict] = None,
        name: str = None,
        description: str = None,
        enabled: bool = True,
        func: Union[Callable, str] = None,
        **func_kwargs,
    ):
        """Create a new stop condition object.

        Args:
            to_check (List[dict], optional): A list of dicts, that holds the
                functions to call and the arguments to call them with. The only
                requirement for the dict is that the ``func`` key is available.
                All other keys are unpacked and passed as kwargs to the given
                function. The ``func`` key can be either a callable or a string
                corresponding to a name in the utopya.stopcond_funcs module.
            name (str, optional): The name of this stop condition
            description (str, optional): A short description of this stop
                condition
            enabled (bool, optional): Whether this stop condition should be
                checked; if False, it will be created but will always be un-
                fulfilled when checked.
            func (Union[Callable, str], optional): (For the short syntax
                only!) If no ``to_check`` argument is given, a function can be
                given here that will be the only one that is checked. If this
                argument is a string, it is also resolved from the utopya
                stopcond_funcs module.
            **func_kwargs: (For the short syntax) The kwargs that are passed
                to the single stop condition function
        """
        self.to_check = self._resolve_sc_funcs(to_check, func, func_kwargs)

        self.enabled = enabled
        self.description = description
        self.name = (
            name
            if name
            else " && ".join([fspec[1] for fspec in self.to_check])
        )

        # Keep track of tasks this stop condition was fulfilled for
        self._fulfilled_for = set()

        # Store the initialization kwargs such that they can be used for yaml
        # representation
        self._init_kwargs = dict(
            to_check=to_check,
            name=name,
            description=description,
            enabled=enabled,
            func=func,
            **func_kwargs,
        )

        log.debug(
            "Initialized stop condition '%s' with %d checking " "function(s).",
            self.name,
            len(self.to_check),
        )

    @property
    def fulfilled_for(self) -> Set["utopya.task.Task"]:
        """The set of tasks this stop condition was fulfilled for"""
        return self._fulfilled_for

    @staticmethod
    def _resolve_sc_funcs(
        to_check: List[dict], func: Union[str, Callable], func_kwargs: dict
    ) -> List[tuple]:
        """Resolves the functions and kwargs that are to be checked.

        The callable is either retrieved from the module-level stop condition
        functions registry or, if the given ``func`` is already a callable,
        that one will be used.
        """

        def retrieve_func(
            func_or_func_name: Union[str, Callable]
        ) -> Tuple[Callable, str]:
            """If this is already a callable, retrieve the name and return that
            tuple. Otherwise retrieve that information from the module-level
            stop condition function registry.
            """
            if callable(func_or_func_name):
                func = func_or_func_name
                return func, func.__name__

            elif not isinstance(func_or_func_name, str):
                raise TypeError(
                    "Expected callable or name of a registered stop condition "
                    f"function, but got {type(func_or_func_name).__name__} "
                    f"with value {repr(func_or_func_name)}!"
                )

            func_name = func_or_func_name
            log.debug(
                "Getting function with name '%s' from the registry.", func_name
            )
            func = STOP_CONDITION_FUNCS.get(func_name)

            if not func or not callable(func):
                _avail = ", ".join(STOP_CONDITION_FUNCS.keys())
                raise ValueError(
                    f"No stop condition function '{func_name}' available! "
                    f"Registered functions: {_avail}"
                )

            return func, func_name

        # Check different argument combinations . . . . . . . . . . . . . . . .
        # Simple case: Without `to_check`
        if func and not to_check:
            log.debug(
                "Got `func` directly and no `to_check` argument; will "
                "use only this function for checking."
            )
            func, func_name = retrieve_func(func)

            return [(func, func_name, func_kwargs)]

        elif to_check and (func or func_kwargs):
            raise ValueError(
                "Got arguments `to_check` and (one or more of) "
                "`func` or `func_kwargs`! Please pass either the "
                "`to_check` (list of dicts) or a single `func` "
                "with a dict of `func_kwargs`."
            )

        elif to_check is None and func is None:
            raise TypeError(
                "Need at least one of the required "
                "keyword-arguments `to_check` or `func`!"
            )

        # Multiple functions: need to resolve the `to_check` list
        funcs_and_kws = []

        for func_dict in to_check:
            # Work on a copy (to be able to pop the `func` entry off and reduce
            # mutability issues in the remaining dict: the func_kwargs)
            func_dict = copy.deepcopy(func_dict)
            func, func_name = retrieve_func(func_dict.pop("func"))
            log.debug("Got function '%s' for stop condition ...", func_name)

            funcs_and_kws.append((func, func_name, func_dict))

        log.debug(
            "Resolved %d stop condition function(s).", len(funcs_and_kws)
        )
        return funcs_and_kws

    def __str__(self) -> str:
        """A string representation for this StopCondition, including the name
        and, if given, the description.
        """
        if self.description:
            return f"StopCondition '{self.name}': {self.description}"
        return f"StopCondition '{self.name}'"

    def fulfilled(self, task: "utopya.task.Task") -> bool:
        """Checks if the stop condition is fulfilled for the given worker,
        using the information from the dict.

        All given stop condition functions are evaluated; if all of them return
        True, this method will also return True.

        Furthermore, if the stop condition is fulfilled, the task's set of
        fulfilled stop conditions will

        Args:
            task (utopya.task.Task): Task object that is to be checked

        Returns:
            bool: If all stop condition functions returned true for the given
                worker and its current information
        """
        if not self.enabled or not self.to_check:
            return False

        # Now perform the check on this task with all stop condition functions
        for sc_func, name, kws in self.to_check:
            if not sc_func(task, **kws):
                # One was not True -> not fulfilled, need not check the others
                log.debug("%s not fulfilled! Returning False ...", name)
                return False

        # All were True -> fulfilled
        task.fulfilled_stop_conditions.add(self)
        self._fulfilled_for.add(task)
        return True

    # YAML Constructor & Representer ..........................................
    yaml_tag = "!stop-condition"

    @classmethod
    def to_yaml(cls, representer, node):
        """Creates a yaml representation of the StopCondition object by storing
        the initialization kwargs as a yaml mapping.

        Args:
            representer (ruamel.yaml.representer): The representer module
            node (StopCondition): The node, i.e. an instance of this class

        Returns:
            a yaml mapping that is able to recreate this object
        """
        # Filter out certain entries that are None
        d = copy.deepcopy(node._init_kwargs)
        d = {
            k: v
            for k, v in d.items()
            if not (
                k in ("name", "description", "func", "to_check") and v is None
            )
            and not (k == "enabled" and v is True)
        }

        # Create the mapping representation from the filtered dict
        return representer.represent_mapping(cls.yaml_tag, d)

    @classmethod
    def from_yaml(cls, constructor, node):
        """Creates a StopCondition object by unpacking the given mapping such
        that all stored arguments are available to ``__init__``.
        """
        return cls(**constructor.construct_mapping(node, deep=True))


def stop_condition_function(f: Callable):
    """A decorator that registers the decorated callable in the module-level
    stop condition function registry. The callable's ``__name__`` attribute
    will be used as the key.

    Args:
        f (Callable): A callable that is to be added to the function registry.

    Raises:
        AttributeError: If the name already exists in the registry
    """
    func_name = f.__name__
    if (
        func_name in STOP_CONDITION_FUNCS
        and STOP_CONDITION_FUNCS.get(func_name) is not f
    ):
        raise AttributeError(
            f"A stop condition function with name '{func_name}' is already "
            f"registered, pointing to {STOP_CONDITION_FUNCS[func_name]}! "
            "Please choose a different name for the to-be-registered function."
        )

    STOP_CONDITION_FUNCS[func_name] = f
    return f


# -----------------------------------------------------------------------------
# -- Stop condition functions -------------------------------------------------
# -----------------------------------------------------------------------------


@stop_condition_function
def timeout_wall(task: "utopya.task.WorkerTask", *, seconds: float) -> bool:
    """Checks the wall timeout of the given worker

    Args:
        task (utopya.task.WorkerTask): The WorkerTask object to check
        seconds (float): After how many seconds to trigger the wall timeout

    Returns:
        bool: Whether the timeout is fulfilled
    """
    return bool((time.time() - task.profiling["create_time"]) > seconds)


@stop_condition_function
def check_monitor_entry(
    task: "utopya.task.WorkerTask",
    *,
    entry_name: str,
    operator: str,
    value: float,
) -> bool:
    """Checks if a monitor entry compares in a certain way to a given value

    Args:
        task (utopya.task.WorkerTask): The WorkerTask object to check
        entry_name (str): The name of the monitor entry, leading to the value
            to the left-hand side of the operator
        operator (str): The binary operator to use
        value (float): The right-hand side value to compare to

    Returns:
        bool: Result of op(entry, value)
    """
    # See if there were and parsed objects
    if not task.outstream_objs:
        # Nope. Nothing to check yet.
        return False

    # Try to recursively retrieve the entry from the latest monitoring output
    latest_monitor = task.outstream_objs[-1]
    try:
        entry = _recursive_getitem(latest_monitor, keys=entry_name.split("."))

    except KeyError:
        # Only warn once
        if entry_name not in _FAILED_MONITOR_ENTRY_CHECKS:
            log.caution(
                "Failed evaluating stop condition due to missing entry '%s' "
                "in monitor output!\nAvailable monitor data: %s",
                entry_name,
                latest_monitor,
            )
            _FAILED_MONITOR_ENTRY_CHECKS.append(entry_name)
        return False

    # Now perform the comparison
    return _OPERATORS[operator](entry, value)



"""Custom data operations"""

import copy as _copy

import dantro as _dantro

from .transform import is_operation as _is_operation

# -----------------------------------------------------------------------------


@_is_operation
def update_attrs(d, attrs: dict):
    """Updates the data attributes in ``d`` with ``attrs``.

    Args:
        d (xarray.DataArray): The data array to write the attributes *to*.
        attrs (dict): The attributes to use for updating

    Returns:
        xarray.DataArray:
            A new data array like ``d`` with updated attributes.
    """
    if isinstance(d, _dantro.base.BaseDataContainer):
        d = d.data

    d = d.copy()
    d.attrs.update(_copy.deepcopy(attrs))
    return d


@_is_operation
def update_with_attrs_from(t, s):
    """Updates the data attributes in ``t`` with those from ``s``.

    Args:
        t (xarray.DataArray): The data array to write the attributes *to*.
        s (xarray.DataArray): The data array to copy the attributes *from*.

    Returns:
        xarray.DataArray:
            A new data array for ``t`` with updated attributes from ``s``.
    """
    if isinstance(t, _dantro.base.BaseDataContainer):
        t = t.data
    if isinstance(s, _dantro.base.BaseDataContainer):
        s = s.data

    return update_attrs(t, s.attrs)



"""Tests the utopya test <...> subcommand of the CLI"""

import os
import sys

import pytest

from .._fixtures import *
from . import invoke_cli

# -----------------------------------------------------------------------------


# FIXME for whatever strange reason, this creates side effects in
#       eval/test_plotting.py::test_preloading
#       Removing "model_plots" from sys.modules here does not resolve it.
@pytest.mark.order("last")
def test_test(with_test_models):
    """Tests `utopya test`"""
    # No tests available for dummy model
    res = invoke_cli(("test", DUMMY_MODEL))
    print(res.output)
    assert res.exit_code == 0
    assert DUMMY_MODEL in res.output
    assert "No tests defined for this model" in res.output
    assert "test summary" not in res.output

    res2 = invoke_cli(("test", DUMMY_MODEL, "-xv"))
    assert res2.output == res.output

    # It does work for the advanced model
    res = invoke_cli(("test", ADVANCED_MODEL))
    print(res.output)
    assert res.exit_code == 0
    assert ADVANCED_MODEL in res.output
    assert "Full test command" in res.output
    assert "pytest" in res.output
    assert "test session starts" in res.output
    assert "collected 1 item" in res.output
    assert "test session starts" in res.output
    assert "1 passed" in res.output

    # Can also pass a label
    res = invoke_cli(("test", ADVANCED_MODEL, "--label", TEST_LABEL))
    print(res.output)
    assert res.exit_code == 0
    assert ADVANCED_MODEL in res.output
    assert "test session starts" in res.output

    # And further arguments
    res = invoke_cli(("test", ADVANCED_MODEL, "-x", "-v", "."))
    print(res.output)
    assert res.exit_code == 0
    assert ADVANCED_MODEL in res.output
    assert "pytest -x -v ." in res.output
    assert "test session starts" in res.output

    res = invoke_cli(("test", ADVANCED_MODEL, "-xv", "test_cfg_sets.py"))
    print(res.output)
    assert res.exit_code == 0
    assert ADVANCED_MODEL in res.output
    assert "pytest -xv test_cfg_sets.py" in res.output
    assert "test session starts" in res.output

    # May pass arguments that lead to zero tests being discovered
    res = invoke_cli(("test", ADVANCED_MODEL, "-xv", "test_foo.py"))
    print(res.output)
    assert res.exit_code == 4  # usage error
    assert "collected 0 items" in res.output
    assert "test session starts" in res.output
    assert "no tests ran" in res.output
    assert "file or directory not found" in res.output



"""Implements the ModelInfoBundle class, which holds a single bundle of
information about a model.
"""

import copy
import logging
import os
import time
from typing import Optional, Sequence, Tuple, Union

from ..exceptions import MissingProjectError
from ..tools import load_selected_keys, load_yml, pformat, recursive_update

log = logging.getLogger(__name__)

TIME_FSTR = "%y%m%d-%H%M%S"
"""Format string to use for generating readable time stamps"""

# -----------------------------------------------------------------------------


class ModelInfoBundle:
    """A bundle of model information; behaves like a read-only dict"""

    SRC_DIR_SEARCH_PATHS = {
        "model_info": "{model_name:}_info.yml",
        "default_cfg": "{model_name:}_cfg.yml",
        "default_plots": "{model_name:}_plots.yml",
        "base_plots": "{model_name:}_base_plots.yml",
    }
    """Path keys that can be searched within the source directory.
    Values are format strings that are evaluated with additional information
    being available (``model_name``).
    The resulting paths are interpreted as paths relative to the source
    directory.
    """

    FSTR_SUFFIX = "_fstr"
    """A suffix used to detect format strings in the path parsing routine.

    If a key ends with such a suffix, the corresponding value is assumed to be
    a format string and is evaluated. The resulting path is stored, dropping
    the suffix.
    """

    PATHS_SCHEMA = (
        ("executable", str),
        ("model_info", str),
        ("source_dir", str),
        ("default_cfg", str),
        ("default_plots", str),
        ("base_plots", str),
        ("py_tests_dir", str),
        ("py_plots_dir", str),
        ("mv_model_cfg", str),
    )
    """Schema to use for a bundle's ``paths`` entry"""

    METADATA_SCHEMA = (
        ("version", str),
        ("long_name", str),
        ("description", str),
        ("long_description", str),
        ("license", str),
        ("author", str),
        ("email", str),
        ("website", str),
        ("utopya_compatibility", str),
        ("language", str),
        ("requirements", list),
        ("misc", dict),
    )
    """Schema to use for a bundle's ``metadata`` entry"""

    # .........................................................................

    def __init__(
        self,
        *,
        model_name: str,
        paths: dict,
        metadata: dict = None,
        project_name: str = None,
        eval_after_run: bool = None,
        registration_time: str = None,
        missing_path_action: str = "log",
        extract_model_info: bool = False,
    ):
        """Initialize a ModelInfoBundle

        Args:
            model_name (str): Name of the model this info bundle describes
            paths (dict): A dictionary of paths
            metadata (dict, optional): A dictionary of metadata entries
            project_name (str, optional): The project this model is part of
            eval_after_run (bool, optional): Whether a model run should be
                followed by an evaluation.
            registration_time (str, optional): Timestamp of registration
            missing_path_action (str, optional): Action upon a path in the
                ``paths`` dict that does not exist.
                Can be ``ignore``, ``log``, ``warn``, ``raise``.
            extract_model_info (bool, optional): Whether to use the information
                from a potentially existing file at the ``model_info`` path
                to pre-populate this bundle. Any explicitly given arguments
                take precedence over the information from that file.
        """
        self._model_name = model_name
        self._reg_time = (
            registration_time
            if registration_time is not None
            else time.strftime(TIME_FSTR)
        )

        # Parse paths that were passed as argument here
        paths = self._parse_paths(
            **{k: os.path.expanduser(p) for k, p in paths.items() if p},
            missing_path_action=missing_path_action,
        )

        # If a model info file is known at this point, use it to pre-populate
        # the data dict.
        self._d = dict(
            paths=dict(),
            metadata=dict(),
            project_name=project_name,
            eval_after_run=eval_after_run,
        )

        if extract_model_info and paths.get("model_info"):
            self._d = recursive_update(
                self._d,
                self._load_and_parse_model_info(paths["model_info"]),
            )

        # Now populate the data dict with the explicitly passed arguments,
        # which should take precedence over the information from the model info
        err_msg_fstr = "Failed loading {} info for '{}' model info bundle!"

        load_selected_keys(
            paths,
            add_to=self.paths,
            keys=self.PATHS_SCHEMA,
            err_msg_prefix=err_msg_fstr.format("paths", model_name),
        )

        load_selected_keys(
            (metadata if metadata is not None else {}),
            add_to=self.metadata,
            keys=self.METADATA_SCHEMA,
            err_msg_prefix=err_msg_fstr.format("metadata", model_name),
        )

        log.debug(
            "Created configuration bundle for model '%s'.", self.model_name
        )

    # Magic methods ...........................................................

    def __eq__(self, other) -> bool:
        """Compares equality by comparing the stored configuration. Only if
        another ModelInfoBundle is compared against does the model name also
        take part in the comparison.
        """
        if isinstance(other, ModelInfoBundle):
            return self._d == other._d and self._model_name == other.model_name
        return self._d == other

    # Formatting ..............................................................

    def __str__(self) -> str:
        return "<Info bundle for '{}' model>\n{}\n".format(
            self._model_name, pformat(self._d)
        )

    # General Access ..........................................................

    @property
    def model_name(self) -> str:
        return self._model_name

    @property
    def registration_time(self) -> str:
        """Registration time string of this bundle"""
        return self._reg_time

    @property
    def as_dict(self) -> dict:
        """Returns a deep copy of all bundle data. This does NOT include the
        model name and the registration time."""
        return copy.deepcopy(self._d)

    def __getitem__(self, key: str):
        """Direct access to the full bundle data"""
        return self._d[key]

    @property
    def executable(self) -> Optional[str]:
        """The path to the model executable"""
        return self._d["paths"].get("executable")

    @property
    def paths(self) -> dict:
        """Access to the paths information of the bundle"""
        return self._d["paths"]

    @property
    def metadata(self) -> dict:
        """Access to the metadata information of the bundle"""
        return self._d["metadata"]

    @property
    def project_name(self) -> str:
        """Access to the Utopia project name information of the bundle"""
        return self._d["project_name"]

    @property
    def eval_after_run(self) -> Union[bool, None]:
        """Whether a model run should be followed by the evaluation routine."""
        return self._d["eval_after_run"]

    @property
    def project(self) -> Union[None, dict]:
        """Load the project information corresponding to this project. Will be
        None if no project is associated.
        """
        if not self.project_name:
            return None

        from .. import PROJECTS

        return PROJECTS[self.project_name]

    @property
    def missing_paths(self) -> dict:
        """Returns those paths where os.path.exists did not evaluate to True"""
        return {k: p for k, p in self.paths.items() if not os.path.exists(p)}

    # Helpers .................................................................

    def _load_and_parse_model_info(self, path: str) -> dict:
        """Loads the model info file from the given path and parses it.

        Parsing steps:
            - Remove entries ``model_name`` or ``label`` that are not
              relevant at this point.

        Args:
            path (str): Path to the model info YAML file
        """
        d = load_yml(path)

        for key in ("model_name", "label"):
            d.pop(key, None)

        return d

    def _parse_paths(
        self,
        *,
        missing_path_action: str,
        executable: str = None,
        source_dir: str = None,
        base_executable_dir: str = None,
        base_source_dir: str = None,
        model_info: str = None,
        **more_paths,
    ) -> dict:
        """Given the path arguments, parse them into absolute paths.
        There are the following parsing steps:

            1. Evaluate ``executable`` and ``source_dir``, potentially using
               the ``base_*`` arguments to resolve relative paths.
            2. If a ``model_info`` path is given, the directory of that file
               is used for any *missing* ``base_*`` argument.
            3. If a ``source_dir`` was given, that directory is searched for
               further existing paths using ``SRC_DIR_SEARCH_PATHS``.
            4. Empty entries in ``more_paths`` are skipped
            5. Remaining entries in ``more_paths`` that end with the suffix
               specified by the ``FSTR_SUFFIX`` class attribute are interpreted
               as format strings and resolved by providing information on the
               ``model_name`` and all available ``paths``. Relative paths are
               interpreted as relative to the source directory.
            6. For all paths, it is checked whether they exist. The
               ``missing_path_action`` argument determines what to do if not.
        """
        paths = dict()

        # Make sure the base_* are not None (makes os.path.join call easier)
        base_executable_dir = (
            base_executable_dir if base_executable_dir is not None else ""
        )
        base_source_dir = (
            base_source_dir if base_source_dir is not None else ""
        )

        # If a model info file is given, can use its directory as a base
        # directory to determine the missing base_* arguments
        model_info_dir = ""
        if model_info:
            if not os.path.isabs(model_info):
                raise ValueError(
                    "The path to the model info file needs to be absolute, "
                    f"but was:  {model_info}"
                )
            model_info_dir = os.path.dirname(model_info)

            if not base_source_dir:
                base_source_dir = model_info_dir

            if not base_executable_dir:
                base_executable_dir = model_info_dir

            paths["model_info"] = model_info

        # Evaluate the executable file path
        if executable is not None:
            paths["executable"] = os.path.join(base_executable_dir, executable)

        # Prepare an absolute version of the source directory path
        abs_source_dir_path = None
        if source_dir:
            abs_source_dir_path = os.path.realpath(
                os.path.join(base_source_dir, source_dir)
            )

        # If a source directory is given, store it, then auto-detect some files
        if abs_source_dir_path:
            paths["source_dir"] = abs_source_dir_path

            for key, fname_fstr in self.SRC_DIR_SEARCH_PATHS.items():
                # Build the full file path, making the model name available.
                # If that path points to an existing file or directory, add it.
                fname = fname_fstr.format(model_name=self.model_name)
                fpath = os.path.join(abs_source_dir_path, fname)

                if os.path.exists(fpath):
                    paths[key] = fpath

        # Parse remaining path entries
        for key, path in more_paths.items():
            if path is None:
                continue

            # Evaluate potentially existing format string style paths, adding
            # them to the paths dict *without* the suffix that was used to
            # identify them.
            if key.endswith(self.FSTR_SUFFIX):
                subkey = key[: -len(self.FSTR_SUFFIX)]
                path = path.format(model_name=self.model_name, paths=paths)
                if subkey in paths or not os.path.exists(path):
                    continue

                # Path exists and was not yet added; store it under the subkey
                key = subkey

            # Interpret relative paths as relative to the source directory
            if not os.path.isabs(path):
                # Is relative. Try to resolve it, starting with a path relative
                # to the source directory. As a fallback, can still attempt to
                # interpret it relative to the model info file.
                if key in self.SRC_DIR_SEARCH_PATHS and abs_source_dir_path:
                    path = os.path.join(abs_source_dir_path, path)

                elif model_info_dir:
                    path = os.path.join(model_info_dir, path)

                else:
                    raise ValueError(
                        f"Given `{key}` path ({path}) was relative, but "
                        "no source directory was specified relative to which "
                        "this path could be interpreted and no model "
                        "information file was available either!"
                    )

            # All good, can now store it. If it was a path that is not to be
            # seen as relative to the source directory, that will lead to an
            # error being thrown below ...
            paths[key] = path

        # Done populating paths dict.
        # Make sure paths are all absolute and exist.
        for key, path in paths.items():
            if not os.path.isabs(path):
                raise ValueError(
                    f"The given `{key}` path ({path}) for info bundle "
                    f"of model '{self.model_name}' was not absolute! "
                    "Please provide only absolute paths (may include ~)."
                )

            if not os.path.exists(path):
                msg = (
                    f"Given '{key}' path for model '{self.model_name}' does "
                    f"not exist: {path}"
                )

                if missing_path_action == "warn":
                    log.warning(msg)

                elif missing_path_action in ("log", "ignore"):
                    log.debug(msg)

                elif missing_path_action == "raise":
                    raise ValueError(
                        msg + "\nEither adjust the corresponding model info "
                        "bundle or place the expected file at that path. "
                        "To ignore this error, pass the `missing_path_action` "
                        "argument to ModelInfoBundle."
                    )

                else:
                    raise ValueError(
                        f"Invalid missing_path_action '{missing_path_action}'!"
                        " Choose from: ignore, log, warn, raise."
                    )

        return paths

    # YAML representation .....................................................

    @classmethod
    def to_yaml(cls, representer, node):
        """Creates a YAML representation of the data stored in this bundle.

        Args:
            representer (ruamel.yaml.representer): The representer module
            node (ModelInfoBundle): The node to represent, i.e. an instance of
                this class.

        Returns:
            A YAML representation of the given instance of this class.
        """
        return representer.represent_data(node._d)



"""Implements utopya-specializations of :py:mod:`dantro.plot.creators`"""

import logging

import dantro
import dantro.plot.creators

from .plothelper import PlotHelper

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


class PyPlotCreator(dantro.plot.creators.PyPlotCreator):
    """This is the Utopia-specific version of dantro's
    :py:class:`~dantro.plot.creators.pyplot.PyPlotCreator`.

    Its main purpose is to define common settings for plotting. By adding this
    extra layer, it allows for future extensibility as well.
    """

    EXTENSIONS = "all"
    """Which file extensions to support.
    A value of ``all`` leads to no checks being performed on the extension.
    """

    DEFAULT_EXT = "pdf"
    """Default plot file extension"""

    PLOT_HELPER_CLS = PlotHelper
    """The PlotHelper class to use; here, the utopya-specific one"""


class UniversePlotCreator(
    dantro.plot.creators.UniversePlotCreator, PyPlotCreator
):
    """Makes plotting with data from a single universe more convenient"""

    PSGRP_PATH: str = "multiverse"
    """The path within the data tree to arrive at the
    :py:class:`~utopya.eval.groups.MultiverseGroup` that this plot creator
    expects universes to be located *in*.
    """


class MultiversePlotCreator(
    dantro.plot.creators.MultiversePlotCreator, PyPlotCreator
):
    """Makes plotting with data from *all* universes more convenient"""

    PSGRP_PATH: str = "multiverse"
    """The path within the data tree to arrive at the
    :py:class:`~utopya.eval.groups.MultiverseGroup`.
    """
